from typing import List

from nipype import config

from clinica.pipelines.engine import Pipeline

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class T1VolumeTissueSegmentation(Pipeline):
    """T1VolumeTissueSegmentation - Tissue segmentation, bias correction and spatial normalization to MNI space.

    Returns:
        A clinica pipeline object containing the T1VolumeTissueSegmentation pipeline.
    """

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        from clinica.utils.spm import get_tpm

        self.parameters.setdefault("tissue_classes", [1, 2, 3])
        self.parameters.setdefault("dartel_tissues", [1, 2, 3])
        self.parameters.setdefault("save_warped_unmodulated", True)
        self.parameters.setdefault("save_warped_modulated", False)
        self.parameters.setdefault("tissue_probability_maps", None)
        # Fetch the TPM in case none was passed explicitly.
        if not self.parameters["tissue_probability_maps"]:
            self.parameters["tissue_probability_maps"] = get_tpm()

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["t1w"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return [
            "bias_corrected_images",
            "bias_field_images",
            "dartel_input_images",
            "forward_deformation_field",
            "inverse_deformation_field",
            "modulated_class_images",
            "native_class_images",
            "normalized_class_images",
            "transformation_mat",
            "t1_mni",
        ]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline.

        Raise:
            ClinicaBIDSError: If there are duplicated files or missing files for any subject
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.iotools.utils.data_handling import (
            check_volume_location_in_world_coordinate_system,
        )
        from clinica.utils.exceptions import ClinicaBIDSError, ClinicaException
        from clinica.utils.input_files import T1W_NII
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        # Inputs from anat/ folder
        # ========================
        # T1w file:
        try:
            t1w_files, _ = clinica_file_reader(
                self.subjects, self.sessions, self.bids_directory, T1W_NII
            )
        except ClinicaException as e:
            err = f"Clinica faced error(s) while trying to read files in your BIDS directory.\n{str(e)}"
            raise ClinicaBIDSError(err)

        check_volume_location_in_world_coordinate_system(
            t1w_files,
            self.bids_directory,
            skip_question=self.parameters["skip_question"],
        )

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint("The pipeline will last approximately 10 minutes per image.")

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("t1w", t1w_files),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (read_node, self.input_node, [("t1w", "t1w")]),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        pass

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.spm as spm
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import unzip_nii, zip_nii
        from clinica.utils.nipype import container_from_filename, fix_join
        from clinica.utils.spm import use_spm_standalone_if_available

        from .t1_volume_tissue_segmentation_utils import (
            ApplySegmentationDeformation,
            get_tissue_tuples,
            init_input_node,
            print_end_pipeline,
            zip_list_files,
        )

        use_spm_standalone_if_available()

        # Get <subject_id> (e.g. sub-CLNC01_ses-M000) from input_node
        # and print begin message
        # =======================
        init_node = npe.Node(
            interface=nutil.Function(
                input_names=self.get_input_fields(),
                output_names=["subject_id"] + self.get_input_fields(),
                function=init_input_node,
            ),
            name="0-InitNode",
        )

        unzip_node = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="1-UnzipT1w",
        )

        # Unified Segmentation
        # ====================
        new_segment = npe.Node(spm.NewSegment(), name="2-SpmSegmentation")
        new_segment.inputs.write_deformation_fields = [True, True]
        new_segment.inputs.tissues = get_tissue_tuples(
            self.parameters["tissue_probability_maps"],
            self.parameters["tissue_classes"],
            self.parameters["dartel_tissues"],
            self.parameters["save_warped_unmodulated"],
            self.parameters["save_warped_modulated"],
        )

        # Apply segmentation deformation to T1 (into MNI space)
        # =====================================================
        t1_to_mni = npe.Node(ApplySegmentationDeformation(), name="3-T1wToMni")

        # Print end message
        # =================
        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=["subject_id", "final_file"],
                function=print_end_pipeline,
            ),
            name="WriteEndMessage",
        )
        self.connect(
            [
                (self.input_node, init_node, [("t1w", "t1w")]),
                (init_node, unzip_node, [("t1w", "in_file")]),
                (unzip_node, new_segment, [("out_file", "channel_files")]),
                (init_node, print_end_message, [("subject_id", "subject_id")]),
                (unzip_node, t1_to_mni, [("out_file", "in_files")]),
                (
                    new_segment,
                    t1_to_mni,
                    [("forward_deformation_field", "deformation_field")],
                ),
                (
                    new_segment,
                    self.output_node,
                    [
                        ("bias_corrected_images", "bias_corrected_images"),
                        ("bias_field_images", "bias_field_images"),
                        ("dartel_input_images", "dartel_input_images"),
                        ("forward_deformation_field", "forward_deformation_field"),
                        ("inverse_deformation_field", "inverse_deformation_field"),
                        ("modulated_class_images", "modulated_class_images"),
                        ("native_class_images", "native_class_images"),
                        ("normalized_class_images", "normalized_class_images"),
                        ("transformation_mat", "transformation_mat"),
                    ],
                ),
                (t1_to_mni, self.output_node, [("out_files", "t1_mni")]),
                (self.output_node, print_end_message, [("t1_mni", "final_file")]),
            ]
        )

        # Find container path from t1w filename
        # =====================================
        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="ContainerPath",
        )

        # Writing CAPS
        # ============
        write_node = npe.Node(name="WriteCAPS", interface=nio.DataSink())
        write_node.inputs.base_directory = str(self.caps_directory)
        write_node.inputs.parameterization = False
        write_node.inputs.regexp_substitutions = [
            (r"(.*)c1(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-graymatter\3"),
            (r"(.*)c2(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-whitematter\3"),
            (r"(.*)c3(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-csf\3"),
            (r"(.*)c4(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-bone\3"),
            (r"(.*)c5(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-softtissue\3"),
            (r"(.*)c6(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-background\3"),
            (r"(.*)(/native_space/sub-.*)(\.nii(\.gz)?)$", r"\1\2_probability\3"),
            (
                r"(.*)(/([a-z]+)_deformation_field/)i?y_(sub-.*)(\.nii(\.gz)?)$",
                r"\1/normalized_space/\4_target-Ixi549Space_transformation-\3_deformation\5",
            ),
            (
                r"(.*)(/t1_mni/)w(sub-.*)_T1w(\.nii(\.gz)?)$",
                r"\1/normalized_space/\3_space-Ixi549Space_T1w\4",
            ),
            (
                r"(.*)(/modulated_normalized/)mw(sub-.*)(\.nii(\.gz)?)$",
                r"\1/normalized_space/\3_space-Ixi549Space_modulated-on_probability\4",
            ),
            (
                r"(.*)(/normalized/)w(sub-.*)(\.nii(\.gz)?)$",
                r"\1/normalized_space/\3_space-Ixi549Space_modulated-off_probability\4",
            ),
            (r"(.*/dartel_input/)r(sub-.*)(\.nii(\.gz)?)$", r"\1\2_dartelinput\3"),
            # Will remove trait_added empty folder
            (r"trait_added", r""),
        ]

        self.connect(
            [
                (self.input_node, container_path, [("t1w", "bids_or_caps_filename")]),
                (
                    container_path,
                    write_node,
                    [
                        (
                            ("container", fix_join, "t1", "spm", "segmentation"),
                            "container",
                        )
                    ],
                ),
                (
                    self.output_node,
                    write_node,
                    [
                        (("native_class_images", zip_list_files, True), "native_space"),
                        (("dartel_input_images", zip_list_files, True), "dartel_input"),
                    ],
                ),
                (
                    self.output_node,
                    write_node,
                    [
                        (
                            ("inverse_deformation_field", zip_nii, True),
                            "inverse_deformation_field",
                        )
                    ],
                ),
                (
                    self.output_node,
                    write_node,
                    [
                        (
                            ("forward_deformation_field", zip_nii, True),
                            "forward_deformation_field",
                        )
                    ],
                ),
                (self.output_node, write_node, [(("t1_mni", zip_nii, True), "t1_mni")]),
            ]
        )
        if self.parameters["save_warped_unmodulated"]:
            self.connect(
                [
                    (
                        self.output_node,
                        write_node,
                        [
                            (
                                ("normalized_class_images", zip_list_files, True),
                                "normalized",
                            )
                        ],
                    ),
                ]
            )
        if self.parameters["save_warped_modulated"]:
            self.connect(
                [
                    (
                        self.output_node,
                        write_node,
                        [
                            (
                                ("modulated_class_images", zip_list_files, True),
                                "modulated_normalized",
                            )
                        ],
                    ),
                ]
            )



import click

from .adni_to_bids import adni_to_bids_cli
from .aibl_to_bids import aibl_to_bids_cli
from .genfi_to_bids import genfi_to_bids_cli
from .habs_to_bids import habs_to_bids_cli
from .ixi_to_bids import ixi_to_bids_cli
from .nifd_to_bids import nifd_to_bids_cli
from .oasis3_to_bids import oasis3_to_bids_cli
from .oasis_to_bids import oasis_to_bids_cli
from .ukb_to_bids import ukb_to_bids_cli


@click.group("convert")
def cli() -> None:
    """Convert popular neuroimaging datasets to the BIDS format."""
    pass


cli.add_command(adni_to_bids_cli.cli)
cli.add_command(aibl_to_bids_cli.cli)
cli.add_command(habs_to_bids_cli.cli)
cli.add_command(nifd_to_bids_cli.cli)
cli.add_command(oasis_to_bids_cli.cli)
cli.add_command(oasis3_to_bids_cli.cli)
cli.add_command(ukb_to_bids_cli.cli)
cli.add_command(genfi_to_bids_cli.cli)
cli.add_command(ixi_to_bids_cli.cli)

if __name__ == "__main__":
    cli()



from pathlib import Path

__all__ = ["compute_atlas", "write_tsv_files"]


def compute_atlas(
    caps_directory: Path,
    to_process_with_atlases: tuple,
    path_to_atlas: Path,
) -> tuple[Path, str, str]:
    import subprocess
    from pathlib import Path
    from tempfile import mkdtemp

    from clinica.compat import errno
    from clinica.utils.stream import cprint

    subject_dir = Path()
    image_id = ""
    atlas = ""

    for path in path_to_atlas.rglob(f"*{to_process_with_atlases[0]}_6p0.gcs"):
        hemisphere = Path(path.stem).stem
        atlas_name = Path(path.stem).suffix[1:].split("_")[0]
        if "long" not in to_process_with_atlases[1]:
            sub, ses = to_process_with_atlases[1].split("_")
            subject_dir = (
                caps_directory
                / "subjects"
                / sub
                / ses
                / "t1"
                / "freesurfer_cross_sectional"
            )
            substitute_dir = subject_dir
            freesurfer_cross = subject_dir / to_process_with_atlases[1]
            output_path_annot = (
                freesurfer_cross / "label" / f"{hemisphere}.{atlas_name}.annot"
            )
            sphere_reg = freesurfer_cross / "surf" / f"{hemisphere}.sphere.reg"
            output_path_stats = (
                freesurfer_cross / "stats" / f"{hemisphere}.{atlas_name}.stats"
            )
            subject_id = to_process_with_atlases[1]
        else:
            sub, ses, long = to_process_with_atlases[1].split("_")
            substitute_dir = Path(mkdtemp())
            substitute_dir.mkdir(exist_ok=True)
            subject_id = f"{sub}_{ses}.long.{sub}_{long}"
            subject_dir = (
                caps_directory
                / "subjects"
                / sub
                / ses
                / "t1"
                / long
                / "freesurfer_longitudinal"
            )
            freesurfer_cross = subject_dir / subject_id
            try:
                (substitute_dir / subjid).symlink_to(freesurfer_cross)
            except FileExistsError as e:
                if e.errno != errno.EEXIST:  # EEXIST: folder already exists
                    raise e
            substitute_dir_id = substitute_dir / subject_id
            output_path_annot = (
                substitute_dir_id / "label" / f"{hemisphere}.{atlas_name}.annot"
            )
            sphere_reg = substitute_dir_id / "surf" / f"{hemisphere}.sphere.reg"
            output_path_stats = (
                substitute_dir_id / "stats" / f"{hemisphere}.{atlas_name}.stats"
            )
        if not sphere_reg.is_file():
            cprint(
                f"The {hemisphere}.sphere.reg file appears to be missing. "
                f"The data for {to_process_with_atlases[1]} will not be "
                f"processed with {to_process_with_atlases[0]}",
                lvl="warning",
            )
        mris_ca_label_command = (
            f"mris_ca_label -sdir {substitute_dir} {subject_id} {hemisphere} "
            f"{freesurfer_cross}/surf/{hemisphere}.sphere.reg {path} {output_path_annot}"
        )
        subprocess.run(mris_ca_label_command, shell=True, capture_output=True)
        mris_anatomical_stats_command = (
            f"export SUBJECTS_DIR={substitute_dir}; "
            f"mris_anatomical_stats -a {output_path_annot} -f {output_path_stats} -b {subject_id} {hemisphere}"
        )
        subprocess.run(mris_anatomical_stats_command, shell=True, capture_output=True)
        image_id, atlas = (
            to_process_with_atlases[1],
            to_process_with_atlases[0],
        )
    return subject_dir, image_id, atlas


def write_tsv_files(subject_dir: Path, image_id: str, atlas: str) -> str:
    """
    Generate statistics TSV files in `subjects_dir`/regional_measures folder for `image_id`.

    Notes
    -----
    We do not need to check the line "finished without error" in scripts/recon-all.log.
    If an error occurs, it will be detected by Nipype and the next nodes (including
    write_tsv_files will not be called).
    """
    from clinica.pipelines.anatomical.freesurfer.utils import generate_regional_measures
    from clinica.utils.stream import cprint

    if "long" in image_id:
        sub, ses, long = image_id.split("_")
        folder = f"{sub}_{ses}.long.{sub}_{long}"
    else:
        folder = image_id
    if (subject_dir / folder / "mri" / "aparc+aseg.mgz").is_file():
        generate_regional_measures(subject_dir, folder, [atlas])
    else:
        cprint(
            msg=(
                f"{image_id.replace('_', ' | ')} does not contain "
                f"mri/aseg+aparc.mgz file. Creation of regional_measures/ folder will be skipped."
            ),
            lvl="warning",
        )
    return image_id



from pathlib import Path
from typing import Optional

__all__ = [
    "init_input_node",
    "run_clinica_surfstat",
    "save_to_caps",
    "get_t1_freesurfer_custom_file",
    "get_pet_surface_custom_file",
    "create_glm_info_dictionary",
    "build_design_matrix",
]


def get_t1_freesurfer_custom_file() -> str:
    import os

    return os.path.join(
        "%(subject)s",
        "%(session)s",
        "t1",
        "freesurfer_cross_sectional",
        "%(subject)s_%(session)s",
        "surf",
        "%(hemi)s.thickness.fwhm%(fwhm)s.fsaverage.mgh",
    )


def get_pet_surface_custom_file(acq_label: str, suvr_reference_region: str) -> str:
    import os

    return os.path.join(
        "%(subject)s",
        "%(session)s",
        "pet",
        "surface",
        f"%(subject)s_%(session)s_trc-{acq_label}_pet"
        f"_space-fsaverage_suvr-{suvr_reference_region}_pvc-iy_hemi-%(hemi)s_fwhm-%(fwhm)s_projection.mgh",
    )


def init_input_node(parameters: dict, base_dir, subjects_visits_tsv):
    """Initialize the pipeline.

    This function will:
        - Create `surfstat_results_dir` in `base_dir`/<group_id> for SurfStat;
        - Save pipeline parameters in JSON file;
        - Copy TSV file with covariates;
        - Print begin execution message.

    Parameters
    ----------
    parameters : dict
        The pipeline's parameters.

    base_dir : Path
        The path to the pipeline's base directory.
        This is a pathlib Path. No type hints because of Nipype.

    subjects_visits_tsv : Path
        The path to the subjects TSV file.
        This is a pathlib Path. No type hints because of Nipype.

    Returns
    -------
    group_label : str
        The group label.

    surfstat_results_dir : Path
        The folder which will contain the results for SurfStat.
    """
    import json
    import shutil

    from clinica.pipelines.statistics_surface._utils import create_glm_info_dictionary
    from clinica.utils.ux import print_begin_image

    group_id = "group-" + parameters["group_label"]
    surfstat_results_dir = base_dir / group_id
    surfstat_results_dir.mkdir(parents=True, exist_ok=True)

    # Save pipeline parameters in JSON file
    glm_dict = create_glm_info_dictionary(subjects_visits_tsv, parameters)
    with open(surfstat_results_dir / f"{group_id}_glm.json", "w") as json_file:
        json.dump(glm_dict, json_file, indent=4)

    # Copy TSV file with covariates
    tsv_filename = surfstat_results_dir / f"{group_id}_covariates.tsv"
    shutil.copyfile(subjects_visits_tsv, tsv_filename)

    # Print begin message
    list_keys = ["AnalysisType", "Covariates", "Contrast", "FWHM", "ClusterThreshold"]
    list_values = [
        parameters["glm_type"],
        parameters["covariates"],
        parameters["contrast"],
        str(parameters["full_width_at_half_maximum"]),
        str(parameters["cluster_threshold"]),
    ]
    group_id = "group-" + parameters["group_label"]
    print_begin_image(group_id, list_keys, list_values)

    return parameters["group_label"], surfstat_results_dir


def _get_string_format_from_tsv(tsv_file: Path) -> str:
    """Determine string format from TSV file.

    If the TSV file is like:

    participant_id  session_id  sex     group   age
    sub-CLNC0001    ses-M000     Female  CN      71.1
    sub-CLNC0002    ses-M000     Male    CN      81.3
    sub-CLNC0003    ses-M000     Male    CN      75.4

    The columns of the TSV file contains consecutively strings, strings,
    strings, strings and float. The string_format is therefore "%s %s %s %s %f".

    Parameters
    ----------
    tsv_file : Path
        The path to the TSV file.

    Returns
    -------
    str :
        The string formatting of the TSV file (e.g. "%s %s %s %s %f")
    """
    import pandas as pd

    demographics_df = pd.read_csv(tsv_file, sep="\t")

    return " ".join(
        [
            _convert_dtype_to_str_format(demographics_df[column].dtype)
            for column in demographics_df.columns
        ]
    )


def _convert_dtype_to_str_format(dtype) -> str:
    """Convert pandas dtypes (e.g. int64) to string format (e.g. %d)"""
    import numpy as np

    if dtype == np.int64:
        return "%d"
    if dtype == np.float64:
        return "%f"
    if dtype == np.object_:
        return "%s"
    raise ValueError(f"Unknown dtype (given: {dtype})")


def build_design_matrix(contrast: str, covariates: Optional[str] = None) -> str:
    """Generate the design matrix for SurfStat based on the contrast and the optional list of covariates.

    Design matrix "1 + <contrast> + <covariate_1> + ... + <covariate_n>"

    Example
    -------
    >>> from clinica.pipelines.statistics_surface.statistics_surface_utils import _build_design_matrix
    >>> _build_design_matrix('group', 'age sex group')
    1 + group + age + sex
    >>> _build_design_matrix('group', 'age')
    1 + group + age
    >>> _build_design_matrix('group', None)
    1 + group
    """
    if covariates:
        # Convert string to list while handling case where several spaces are present
        list_covariates = list(covariates)
        try:
            list_covariates.remove("")
        except ValueError:
            pass
        if contrast in list_covariates:
            return "1 + " + " + ".join(covariate for covariate in list_covariates)
        return (
            "1 + "
            + contrast
            + " + "
            + " + ".join(covariate for covariate in list_covariates)
        )
    return "1 + " + contrast


def run_clinica_surfstat(
    caps_dir,  # Path. no type hint because of Nipype self-contained requirement
    output_dir,  # Path. no type hint because of Nipype self-contained requirement
    subjects_visits_tsv,  # Path. no type hint because of Nipype self-contained requirement
    pipeline_parameters: dict,
):
    """Call clinica_surfstat function.

    Parameters
    ----------
    caps_dir : Path
        The path to CAPS directory containing surface-based features.

    output_dir : Path
        The path to output directory that will contain outputs.

    subjects_visits_tsv : Path
        The path to TSV file containing the GLM information.

    pipeline_parameters : dict
        Parameters of StatisticsSurface pipeline.

    Returns
    -------
    output_dir : Path
        The path to the output directory.
    """
    from clinica.pipelines.statistics_surface._utils import build_design_matrix
    from clinica.pipelines.statistics_surface.surfstat import clinica_surfstat
    from clinica.utils.check_dependency import get_freesurfer_home

    clinica_surfstat(
        caps_dir / "subjects",
        output_dir,
        subjects_visits_tsv,
        build_design_matrix(
            pipeline_parameters["contrast"], pipeline_parameters["covariates"]
        ),
        pipeline_parameters["contrast"],
        pipeline_parameters["glm_type"],
        pipeline_parameters["group_label"],
        get_freesurfer_home(),
        pipeline_parameters["measure_label"],
        surface_file=pipeline_parameters["custom_file"],
        fwhm=pipeline_parameters["full_width_at_half_maximum"],
        cluster_threshold=pipeline_parameters["cluster_threshold"],
    )
    return output_dir


def create_glm_info_dictionary(tsv_file: Path, pipeline_parameters: dict) -> dict:
    """Create dictionary containing the GLM information that will be stored in a JSON file."""
    glm_info = {
        # Clinica compulsory arguments
        "AnalysisType": pipeline_parameters["glm_type"],
        "DesignMatrix": build_design_matrix(
            pipeline_parameters["contrast"],
            pipeline_parameters["covariates"],
        ),
        "StringFormatTSV": _get_string_format_from_tsv(tsv_file),
        "Contrast": pipeline_parameters["contrast"],
        "GroupLabel": pipeline_parameters["group_label"],
        # Optional arguments
        "Covariates": pipeline_parameters["covariates"],
        "FWHM": pipeline_parameters["full_width_at_half_maximum"],
        # Optional arguments for custom pipeline
        "custom_file": pipeline_parameters["custom_file"],
        "measure_label": pipeline_parameters["measure_label"],
        # Advanced arguments (i.e. tricky parameters)
        "ThresholdUncorrectedPvalue": 0.001,
        "ThresholdCorrectedPvalue": 0.05,
        "ClusterThreshold": pipeline_parameters["cluster_threshold"],
    }
    # Optional arguments for inputs from pet-surface pipeline
    if (
        pipeline_parameters["acq_label"]
        and pipeline_parameters["suvr_reference_region"]
    ):
        glm_info["acq_label"] = pipeline_parameters["acq_label"]
        glm_info["suvr_reference_region"] = pipeline_parameters["suvr_reference_region"]

    return glm_info


def save_to_caps(
    source_dir,  # Path. no type hint because of Nipype self-contained requirement
    caps_dir,  # Path. no type hint because of Nipype self-contained requirement
    overwrite_caps: bool,
    group_label: str,
    glm_type: str,
) -> None:
    """Save `source_dir`/ to CAPS folder.

    This function copies outputs of `source_dir`/ to
    `caps_dir`/groups/<group_id>/<statistics>/surfstat_<glm_type>/

    The `source_dir`/ folder should contain the following elements:
        - group-<group_label>_<group_1_or_2>-lt-<group_1_or_2>_measure-<measure>_fwhm-<label>_suffix.ext
    or
        - group-<group_label>_correlation-<label>_contrast-{-|+}_measure-<measure>_fwhm-<label>_suffix.ext
    and
        - group-<group_label>_covariates.tsv
        - group-<group_label>_glm.json

    Raises
    ------
    NotImplementedError
        If overwrite_caps=True.
    """
    import shutil

    from clinica.pipelines.statistics_surface.surfstat.models import GLMModelType
    from clinica.utils.ux import print_end_image

    glm_type = GLMModelType(glm_type)
    if glm_type == GLMModelType.GROUP_COMPARISON:
        surfstat_folder = f"surfstat_{glm_type}"
    if glm_type == GLMModelType.CORRELATION:
        surfstat_folder = f"surfstat_{glm_type}_analysis"

    destination_dir = (
        caps_dir.expanduser()
        / "groups"
        / f"group-{group_label}"
        / "statistics"
        / surfstat_folder
    )

    if overwrite_caps:
        raise NotImplementedError("save_to_caps(overwrite_caps=True) not implemented")
    shutil.copytree(source_dir, destination_dir, symlinks=True)
    print_end_image(f"group-{group_label}")



"""This module contains utilities common to DWI preprocessing pipelines."""

from os import PathLike
from pathlib import Path
from typing import Optional, Tuple, Union

import nibabel as nib
import numpy as np

from ..utils import DWIDataset

__all__ = [
    "generate_index_file",
    "generate_acq_file",
    "compute_average_b0",
    "check_file",
    "add_suffix_to_filename",
    "get_b0_filter",
    "check_b_value_threshold",
    "get_readout_time_and_phase_encoding_direction",
    "check_dwi_volume",
    "check_dwi_dataset",
]


def generate_index_file(
    b_values_filename: Path,
    image_id: Optional[str] = None,
    output_dir: Optional[Path] = None,
) -> Path:
    """Generate [`image_id`]_index.txt file for FSL eddy command.

    At the moment, all volumes are assumed to be acquired with the
    same parameters. The generate_acq_file function writes a single
    line, and this function writes a vector of ones linking each
    DWI volume to this first line.

    Parameters
    ----------
    b_values_filename : Path
        The path to the b-values file.

    image_id : str, optional
        An optional prefix for the output file name.
        Defaults to None.

    output_dir : Path, optional
        The path to the directory in which the index file
        should be written. If not provided, it will be written
        in the same folder as the provided b values filename.

    Returns
    -------
    index_filename: Path
        The path to output index file. [`image_id`]_index.txt or index.txt file.
    """
    if not b_values_filename.is_file():
        raise FileNotFoundError(f"Unable to find b-values file: {b_values_filename}.")

    b_values = np.loadtxt(b_values_filename, ndmin=1)
    index_filename = f"{image_id}_index.txt" if image_id else "index.txt"
    output_dir = output_dir or b_values_filename.parent
    index_filename = output_dir / index_filename
    np.savetxt(index_filename, np.ones(len(b_values), dtype=int).T)

    return index_filename


def generate_acq_file(
    dwi_filename: Path,
    fsl_phase_encoding_direction: str,
    total_readout_time: str,
    image_id: Optional[str] = None,
    output_dir: Optional[Path] = None,
) -> Path:
    """Generate [`image_id`]_acq.txt file for FSL eddy command.

    Parameters
    ----------
    dwi_filename : str
        Path to the DWI file.

    fsl_phase_encoding_direction : str
        Phase encoding direction from the BIDS specifications in FSL format
        (i.e. x/y/z instead of i/j/k).

    total_readout_time : str
        Total readout time from BIDS specifications.

    image_id : str, optional
        The prefix for the output file. Defaults to None.

    output_dir : Path, optional
        The path to the directory in which the acquisition file
        should be written. If not provided, it will be written
        in the same folder as the provided dwi filename.

    Returns
    -------
    acq_filename : Path
        The path to the acquisition file.
    """
    if fsl_phase_encoding_direction not in ("x", "y", "z", "x-", "y-", "z-"):
        raise RuntimeError(
            f"FSL PhaseEncodingDirection (found value: {fsl_phase_encoding_direction}) "
            f"is unknown, it should be a value in (x, y, z, x-, y-, z-)"
        )
    acq_filename = f"{image_id}_acq.txt" if image_id else "acq.txt"
    output_dir = output_dir or dwi_filename.parent
    acq_filename = output_dir / acq_filename
    basis_vector = _get_phase_basis_vector(fsl_phase_encoding_direction)
    basis_vector.append(float(total_readout_time))
    np.savetxt(acq_filename, np.array([basis_vector]), fmt="%d " * 3 + "%f")

    return acq_filename


def _get_phase_basis_vector(phase: str) -> list:
    """Returns the unit vector corresponding to the given phase."""
    mult = -1 if phase.endswith("-") else 1
    idx = ["x", "y", "z"].index(phase[0])
    result = [0] * 3
    result[idx] = mult

    return result


def compute_average_b0(
    dwi_filename: PathLike,
    b_value_filename: Optional[PathLike] = None,
    b_value_threshold: float = 5.0,
    squeeze: bool = False,
    out_file: Optional[str] = None,
) -> PathLike:
    """Computes the average of the b0 volumes from DWI dataset.

    Parameters
    ----------
    dwi_filename : PathLike
        The path to the DWI files containing the volumes of interest.

    b_value_filename : PathLike, optional
        The path to the b-values file. This will be used to filter the volumes.
        Only volumes with a corresponding b-value smaller than b_value_threshold
        will be kept for computations. If None, all volumes are kept.
        Default is None.

    b_value_threshold : float, optional
        The threshold to apply to b-values to get the volumes which should be kept.
        Default=5.0.

    squeeze : bool, optional
        Whether the returned volume should be squeezed or not.
        If True, it will be a 3D array, if False, a 4D array with a dummy 4th dimension.
        Default=False

    out_file : str, optional
        Name of the output file.
        If None, the output file will be built from the input DWI file base with
        the suffix '_avg_b0.nii[.gz]'.
        Defaults to None.

    Returns
    -------
    out_file : str
        The path to the nifti image file containing the mean of the b0 volumes.

    Raises
    ------
    FileNotFoundError:
        If the DWI input file does not exist.

    ValueError:
        If b_value_threshold < 0.
    """
    from clinica.utils.image import compute_aggregated_volume, get_new_image_like

    dwi_filename = check_file(dwi_filename)
    check_b_value_threshold(b_value_threshold)
    out_file = out_file or add_suffix_to_filename(dwi_filename, "avg_b0")
    volume_filter = (
        get_b0_filter(b_value_filename, b_value_threshold=b_value_threshold)
        if b_value_filename
        else None
    )
    b0 = compute_aggregated_volume(dwi_filename, np.average, volume_filter)
    if not squeeze:
        b0 = b0[..., np.newaxis]
    b0_img = get_new_image_like(dwi_filename, b0)
    b0_img.to_filename(out_file)

    return out_file


def check_file(filename: Union[str, PathLike], resolve: bool = False) -> Path:
    """Check that filename exists and return a Path object."""
    filename = Path(filename).resolve() if resolve else Path(filename)
    if not filename.exists():
        raise FileNotFoundError(f"File not found : {filename}.")

    return filename


def add_suffix_to_filename(filename: Path, suffix: str) -> Path:
    """Adds the provided suffix to the given file name."""
    ext = filename.suffix
    if ext == ".gz":
        ext = "." + filename.stem.split(".")[-1] + ext
    return filename.parent / f"{filename.name.replace(ext, '')}_{suffix}{ext}"


def get_b0_filter(
    b_value_filename: PathLike, b_value_threshold: float = 5.0
) -> np.ndarray:
    """Return the index for the volumes where b<=low_bval.

    Parameters
    ----------
    b_value_filename : PathLike
        Path to the bval file.

    b_value_threshold : float, optional
        Defines the threshold for the b0 volumes as all volumes which
        satisfy b-value <= b_value_threshold.
        Defaults to 5.0.

    Returns
    -------
    np.ndarray :
        Index of the b0 volumes.

    Raises
    ------
    FileNotFoundError:
        If b_value_filename file cannot be found.
    """
    b_value_filename = check_file(b_value_filename)
    values = np.loadtxt(b_value_filename)

    return np.where(values <= b_value_threshold)[0]


def check_b_value_threshold(b_value_threshold: float) -> None:
    if b_value_threshold < 0:
        raise ValueError(
            f"b_value_threshold should be >=0. You provided {b_value_threshold}."
        )


def get_readout_time_and_phase_encoding_direction(
    dwi_json_filename: Union[str, Path],
) -> tuple[str, str]:
    """Extract the readout time and phase encoding direction from the DWI JSON file."""
    from clinica.utils.filemanip import extract_metadata_from_json

    [total_readout_time, phase_encoding_direction] = extract_metadata_from_json(
        dwi_json_filename,
        [
            "TotalReadoutTime",
            "PhaseEncodingDirection",
        ],
        handle_missing_keys=_handle_missing_keys_dwi,
    )
    phase_encoding_direction = _bids_dir_to_fsl_dir(phase_encoding_direction)

    return total_readout_time, phase_encoding_direction


def _handle_missing_keys_dwi(data: dict, missing_keys: set[str]) -> dict:
    """Find alternative fields from the bids/sub-X/ses-Y/dwi/sub-X_ses-Y_dwi.json
    file to replace those which were not found in this very json.


    Parameters
    ----------
    data: dict
        Dictionary containing the json data.

    missing_keys: set of str
        Set of keys that are required and were not found in the json.

    Returns
    -------
    dict:
        Contains the values for the requested fields.
    """
    handlers = {
        "TotalReadoutTime": _handle_missing_total_readout_time,
        "PhaseEncodingDirection": _handle_missing_phase_encoding_direction,
    }
    try:
        return {k: handlers[k](data, missing_keys) for k in missing_keys}
    except KeyError:
        raise ValueError(
            f"Could not recover the missing keys {missing_keys} from JSON file."
        )


def _handle_missing_total_readout_time(data: dict, missing_keys: set) -> float:
    """Find an alternative field in the json to replace the TotalReadoutTime.

    Parameters
    ----------
    data: dict
        Dictionary containing the json data.

    missing_keys: set
        Set of keys that are required and were not found in the json.

    Returns
    -------
    float:
        Value for TotalReadoutTime.
    """
    from clinica.utils.exceptions import ClinicaException

    if "EstimatedTotalReadoutTime" in data:
        return data["EstimatedTotalReadoutTime"]
    if "PhaseEncodingSteps" in data and "PixelBandwidth" in data:
        if data["PixelBandwidth"] != 0:
            return data["PhaseEncodingSteps"] / data["PixelBandwidth"]
        raise ValueError("Pixel Bandwidth value is not valid.")
    raise ClinicaException("Could not recover the TotalReadoutTime from JSON file.")


def _handle_missing_phase_encoding_direction(data: dict, missing_keys: set) -> float:
    """Find an alternative field in the json to replace the PhaseEncodingDirection.

    Parameters
    ----------
    data: dict
        Dictionary containing the json data.

    missing_keys: set
        Set of keys that are required and were not found in the json.

    Returns
    -------
    float:
        Value for PhaseEncodingDirection.
    """
    from clinica.utils.exceptions import ClinicaException

    if "PhaseEncodingAxis" in data:
        return data["PhaseEncodingAxis"] + "+"
    raise ClinicaException(
        "Could not recover the PhaseEncodingDirection from JSON file."
    )


def _bids_dir_to_fsl_dir(bids_dir):
    """Converts BIDS PhaseEncodingDirection parameters (i,j,k,i-,j-,k-) to FSL direction (x,y,z,x-,y-,z-)."""
    fsl_dir = bids_dir.lower()
    if "i" not in fsl_dir and "j" not in fsl_dir and "k" not in fsl_dir:
        raise ValueError(
            f"Unknown PhaseEncodingDirection {fsl_dir}: it should be a value in (i, j, k, i-, j-, k-)"
        )

    return (
        fsl_dir.replace("i", "x").replace("j", "y").replace("k", "z").replace("+", "")
    )


def check_dwi_volume(dwi_dataset: DWIDataset) -> None:
    """Checks the consistency of a given DWIDataset.

    More precisely, checks that the number of DWI volumes,
    the number of B-values, and the number of B-vectors are equal.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        The DWI dataset to check.

    Raises
    ------
    ValueError:
        If the DWI nifti image is not 4D.

    IOError:
        If the b-values and/or b-vectors are all zeros.

        If the number of DWI volumes, the number of B-values,
        and the number of B-vectors are not equal.
    """
    b_values = np.loadtxt(dwi_dataset.b_values, ndmin=1)
    num_b_values = len(b_values)
    b_vectors = np.loadtxt(dwi_dataset.b_vectors, ndmin=2)
    num_b_vectors = b_vectors.shape[-1]
    num_dwi = _load_nifti_at_least_4d(dwi_dataset.dwi, raise_if_3d=True).shape[-1]

    if np.sum(b_values) == 0 or np.sum(b_vectors) == 0:
        raise IOError(
            "The b-values and/or b-vectors have problematic values (all zeros)."
        )
    if not (num_b_values == num_b_vectors == num_dwi):
        raise IOError(
            f"Number of DWIs, b-vals and b-vecs mismatch "
            f"(# DWI = {num_dwi}, # B-vec = {num_b_vectors}, #B-val = {num_b_values}) "
        )


def _load_nifti_at_least_4d(
    image_path: Path, raise_if_3d: bool = False
) -> nib.Nifti1Image:
    image = nib.load(image_path)
    if len(image.shape) == 3:
        if not raise_if_3d:
            data = image.get_fdata()
            image = nib.Nifti1Image(
                np.expand_dims(data, axis=-1), image.affine, image.header
            )
    if len(image.shape) != 4:
        raise ValueError(f"DWI image {image_path} has not a valid number of dimension.")
    return image


def check_dwi_dataset(dwi_dataset: DWIDataset) -> DWIDataset:
    """Checks that provided input files of the DWI dataset exist.

    If they exist, a new dataset object is returned with file paths
    converted to Path objects instead of strings.

    The resulting dataset will contain absolute paths which can be
    required by downstream tools.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        Input DWI dataset to be checked.

    Returns
    -------
    DWIDataset :
        Checked DWI dataset.

    Raises
    ------
    FileNotFoundError
        If one of the files doesn't exist.
    """
    return DWIDataset(*(check_file(f, resolve=True) for f in dwi_dataset))



from typing import Any

import pydra
from nipype.algorithms.misc import Gunzip
from pydra import Workflow
from pydra.tasks.nipype1.utils import Nipype1Task

from clinica.pydra.engine import clinica_io


def _check_pipeline_parameters(parameters: dict) -> dict:
    """Check the parameters passed to the pipeline.
    Parameters
    ----------
    parameters : dict
        Dictionary of parameters to analyze.
    Returns
    -------
    dict :
        Cleaned dictionary of parameters.
    """
    from clinica.utils.group import check_group_label

    if "group_label" not in parameters.keys():
        raise KeyError("Missing compulsory group_label key in pipeline parameter.")
    parameters.setdefault("dartel_tissues", [1, 2, 3])

    check_group_label(parameters["group_label"])

    return parameters


@clinica_io
def t1volume_register_dartel(
    name: str = "t1-volume-register-dartel", parameters: dict = {}
) -> Workflow:
    """Core workflow  for inter-subject registration using Dartel

    Parameters
    ----------
    name : str
        name of pipeline. Default="t1-volume-tissue-segmentation

    parameters : dict
        dictionary of pipeline parameters

    Returns
    -------
    Workflow
        pydra workflow for core functionalities

    """
    import clinica.pipelines.t1_volume_register_dartel.t1_volume_register_dartel_utils as utils
    from clinica.pydra.t1_volume.tasks import (
        task_create_iteration_parameters,
        task_prepare_dartel_input_images,
    )

    parameters = _check_pipeline_parameters(parameters)

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[
            ("_graph_checksums", Any),
            (
                "dartel_input_tissue",
                dict,
                {"tissue_number": parameters["dartel_tissues"]},
                {"mandatory": True},
            ),
            (
                "dartel_iteration_templates",
                dict,
                {
                    "group_label": parameters["group_label"],
                    "i": range(1, 7),
                },
                {"mandatory": True},
            ),
        ],
        bases=(pydra.specs.BaseSpec,),
    )

    workflow = Workflow(
        name,
        input_spec=input_spec,
    )

    workflow.split("dartel_input_tissue")

    workflow.add(
        Nipype1Task(
            name="unzip_dartel_input_tissue",
            interface=Gunzip(),
            in_file=workflow.lzin.dartel_input_tissue,
        )
        .split("in_file")
        .combine("in_file")
    )

    workflow.add(
        Nipype1Task(
            name="unzip_templates",
            interface=Gunzip(),
            in_file=workflow.lzin.dartel_iteration_templates,
        )
        .split("in_file")
        .combine("in_file")
    )

    workflow.add(
        task_prepare_dartel_input_images(
            name="task_prepare_dartel_input_images",
            interface=task_prepare_dartel_input_images,
            nifti_input=workflow.unzip_dartel_input_tissue.lzout.out_file,
        )
    )

    workflow.add(
        task_create_iteration_parameters(
            name="task_create_iteration_parameters",
            template_input=workflow.unzip_templates.lzout.out_file,
        )
    )

    task_dartel_existing_template = Nipype1Task(
        name="task_dartel_existing_template",
        interface=utils.DARTELExistingTemplate(),
    )

    task_dartel_existing_template.inputs.image_files = (
        workflow.task_prepare_dartel_input_images.lzout.image_files
    )
    task_dartel_existing_template.inputs.iteration_parameters = (
        workflow.task_create_iteration_parameters.lzout.iteration_parameters
    )

    workflow.add(task_dartel_existing_template)

    workflow.set_output(
        [
            (
                "dartel_flow_fields",
                workflow.task_dartel_existing_template.lzout.dartel_flow_fields,
            )
        ]
    )

    return workflow



from typing import List, Optional, Union

import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.pet_volume.pipeline as pydra_pet_volume
from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pydra-pet-volume"


@clinica_pipeline
@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@cli_param.argument.acq_label
@cli_param.argument.suvr_reference_region
@cli_param.option_group.pipeline_specific_options
@cli_param.option.pvc_psf_tsv
@cli_param.option_group.advanced_pipeline_options
@cli_param.option_group.option(
    "-mask",
    "--mask_tissues",
    multiple=True,
    type=click.IntRange(1, 7),
    default=[1, 2, 3],
    help=(
        "Tissue classes (1: gray matter (GM), 2: white matter (WM), 3: cerebrospinal "
        "fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to use "
        "for masking the PET image (default: GM, WM and CSF i.e. --mask_tissues 1 2 3)."
    ),
)
@cli_param.option_group.option(
    "-threshold",
    "--mask_threshold",
    default=0.3,
    show_default=True,
    help="Value used as threshold to binarize the tissue maps.",
)
@cli_param.option_group.option(
    "-pvc_mask",
    "--pvc_mask_tissues",
    multiple=True,
    type=click.IntRange(1, 7),
    default=[1, 2, 3],
    help=(
        "Tissue classes (1: gray matter (GM), 2: white matter (WM), 3: cerebrospinal "
        "fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to use "
        "as mask for PVC (default: GM, WM and CSF i.e. --pvc_mask_tissues 1 2 3)."
    ),
)
@cli_param.option.smooth
@option.global_option_group
@option.n_procs
def cli(
    bids_directory: str,
    caps_directory: str,
    group_label: str,
    acq_label: str,
    suvr_reference_region: Optional[str] = None,
    pvc_psf_tsv: Optional[str] = None,
    mask_tissues: List[int] = (1, 2, 3),
    mask_threshold: float = 0.3,
    pvc_mask_tissues: List[int] = (1, 2, 3),
    smooth: Union[float, List[float], List[List[float]]] = 8.0,
    n_procs: Optional[int] = None,
) -> None:
    """SPM-based pre-processing of PET images (Pydra engine).

    GROUP_LABEL is an user-defined identifier to target a specific group of subjects.

    This pipeline is associated to the DARTEL template that you had created when
    running the t1-volume pipeline.

    ACQ_LABEL corresponds to the label given to the PET acquisition, specifying the
    tracer used.
    Frequently used values are '18FFDG' or '18FAV45'.

    The reference region must be specified to perform intensity normalization.
    Accepted values include: 'pons', 'cerebellumPons', 'pons2', 'cerebellumPons2'.

    Prerequisite: You need to have performed the t1-volume pipeline on your T1-weighted MR images.

    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/PET_Volume/
    """
    parameters = {
        "group_label": group_label,
        "acq_label": acq_label,
        "suvr_reference_region": suvr_reference_region,
        "pvc_psf_tsv": pvc_psf_tsv,
        "mask_tissues": mask_tissues,
        "mask_threshold": mask_threshold,
        "pvc_mask_tissues": pvc_mask_tissues,
        "smooth": smooth,
    }
    pipeline = pydra_pet_volume.build_core_workflow(
        name="pet-volume-pydra",
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters=parameters,
    )
    pydra_utils.run(pipeline, n_procs=n_procs)


if __name__ == "__main__":
    cli()



"""Convert the OASIS3 dataset into BIDS."""

from typing import Optional

from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    path_to_clinical: UserProvidedPath,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
):
    """Convert the entire dataset in BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.
    """
    from clinica.iotools.bids_utils import StudyName, write_modality_agnostic_files
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path
    from .oasis3_utils import (
        dataset_to_bids,
        intersect_data,
        read_clinical_data,
        read_imaging_data,
        write_bids,
    )

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    path_to_clinical = validate_input_path(path_to_clinical)
    if subjects:
        cprint(
            (
                f"Subject filtering is not yet implemented in {get_converter_name(StudyName.OASIS3)} converter. "
                "All subjects available will be converted."
            ),
            lvl="warning",
        )
    if n_procs != 1:
        cprint(
            f"{get_converter_name(StudyName.OASIS3)} converter does not support multiprocessing yet. n_procs set to 1.",
            lvl="warning",
        )
    dict_df = read_clinical_data(path_to_clinical)
    imaging_data = read_imaging_data(path_to_dataset)
    imaging_data, df_small = intersect_data(imaging_data, dict_df)
    participants, sessions, scans = dataset_to_bids(imaging_data, df_small)
    write_bids(
        to=bids_dir,
        participants=participants,
        sessions=sessions,
        scans=scans,
        dataset_directory=path_to_dataset,
    )
    readme_data = {
        "link": "https://sites.wustl.edu/oasisbrains/#access",
        "desc": (
            "OASIS-3 is a retrospective compilation of data for 1378 participants that were collected across several "
            "ongoing projects through the WUSTL Knight ADRC over the course of 30years. Participants include 755 "
            "cognitively normal adults and 622 individuals at various stages of cognitive decline ranging in age from "
            "42-95yrs. All participants were assigned a new random identifier and all dates were removed and "
            "normalized to reflect days from entry into study. The dataset contains 2842 MR sessions which include "
            "T1w, T2w, FLAIR, ASL, SWI, time of flight, resting-state BOLD, and DTI sequences. Many of the MR sessions "
            "are accompanied by volumetric segmentation files produced through FreeSurfer processing. PET imaging from "
            "different tracers, PIB, AV45, and FDG, totaling over 2157 raw imaging scans and the accompanying "
            "post-processed files from the Pet Unified Pipeline (PUP) are also available in OASIS-3."
        ),
    }
    write_modality_agnostic_files(
        study_name=StudyName.OASIS3,
        readme_data=readme_data,
        bids_dir=bids_dir,
    )
    cprint("Conversion to BIDS succeeded.", lvl="info")



from pathlib import Path
from typing import List, Optional

import pandas as pd

__all__ = [
    "create_participants_tsv_file",
    "create_scans_tsv_file",
    "create_sessions_tsv_file",
]


def create_participants_tsv_file(
    input_path: Path,
    clinical_specifications_folder: Path,
    clinical_data_dir: Path,
    delete_non_bids_info: bool = True,
) -> None:
    """Create a participants TSV file for the AIBL dataset where information
    regarding the patients are reported.

    Parameters
    ----------
    input_path : Path
        The path to the input directory.

    clinical_specifications_folder : Path
        The path to the folder containing the clinical specification files.

    clinical_data_dir : Path
        The path to the directory to the clinical data files.

    delete_non_bids_info : bool, optional
        If True delete all the rows of the subjects that are not
        available in the BIDS dataset.
        Default=True.
    """
    import glob
    import os
    from os import path

    import numpy as np

    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    fields_bids = ["participant_id"]
    fields_dataset = []
    prev_location = ""
    prev_sheet = ""
    index_to_drop = []

    specifications = _load_specifications(
        clinical_specifications_folder, "participant.tsv"
    )
    participant_fields_db = specifications[StudyName.AIBL.value]
    field_location = specifications[f"{StudyName.AIBL.value} location"]
    participant_fields_bids = specifications["BIDS CLINICA"]

    # Extract the list of the available fields for the dataset (and the corresponding BIDS version)
    for i in range(0, len(participant_fields_db)):
        if not pd.isnull(participant_fields_db[i]):
            fields_bids.append(participant_fields_bids[i])
            fields_dataset.append(participant_fields_db[i])

    # Init the dataframe that will be saved in the file participant.tsv
    participant_df = pd.DataFrame(columns=fields_bids)

    for i in range(0, len(participant_fields_db)):
        # If a field not empty is found
        if not pd.isnull(participant_fields_db[i]):
            # Extract the file location of the field and read the value from the file
            tmp = field_location[i].split("/")
            location = tmp[0]
            # If a sheet is available
            sheet = tmp[1] if len(tmp) > 1 else ""
            # Check if the file to open for a certain field it's the same of the previous field
            if location == prev_location and sheet == prev_sheet:
                pass
            else:
                file_ext = os.path.splitext(location)[1]
                file_to_read_path = path.join(clinical_data_dir, location)

                if file_ext == ".xlsx":
                    file_to_read = pd.read_excel(
                        glob.glob(file_to_read_path)[0], sheet_name=sheet
                    )
                elif file_ext == ".csv":
                    file_to_read = pd.read_csv(glob.glob(file_to_read_path)[0])
                prev_location = location
                prev_sheet = sheet

            field_col_values = []
            # For each field in fields_dataset extract all the column values
            for j in range(0, len(file_to_read)):
                # Convert the alternative_id_1 to string if is an integer/float
                if participant_fields_bids[i] == "alternative_id_1" and (
                    file_to_read[participant_fields_db[i]].dtype == np.float64
                    or file_to_read[participant_fields_db[i]].dtype == np.int64
                ):
                    if not pd.isnull(file_to_read.at[j, participant_fields_db[i]]):
                        # value_to_append = str(file_to_read.get_value(j, participant_fields_db[i])).rstrip('.0')
                        value_to_append = str(
                            file_to_read.at[j, participant_fields_db[i]]
                        )
                    else:
                        value_to_append = "n/a"
                else:
                    value_to_append = file_to_read.at[j, participant_fields_db[i]]
                field_col_values.append(value_to_append)
            # Add the extracted column to the participant_df
            participant_df[participant_fields_bids[i]] = pd.Series(field_col_values)

    # Compute BIDS-compatible participant ID.
    participant_df["participant_id"] = participant_df["alternative_id_1"].apply(
        lambda x: bids_id_factory(StudyName.AIBL).from_original_study_id(x)
    )
    # Keep year-of-birth only.
    participant_df["date_of_birth"] = participant_df["date_of_birth"].str.extract(
        r"/(\d{4}).*"
    )
    # Normalize sex value.
    participant_df["sex"] = participant_df["sex"].map({1: "M", 2: "F"}).fillna("n/a")

    # Normalize known NA values.
    participant_df.replace(-4, "n/a", inplace=True)

    # Delete all the rows of the subjects that are not available in the BIDS dataset
    if delete_non_bids_info:
        participant_df = participant_df.drop(index_to_drop)

    participant_df.to_csv(
        input_path / "participants.tsv",
        sep="\t",
        index=False,
        encoding="utf8",
    )


def _load_specifications(
    clinical_specifications_folder: Path, filename: str
) -> pd.DataFrame:
    specifications = clinical_specifications_folder / filename
    if not specifications.exists():
        raise FileNotFoundError(
            f"The specifications for {filename} were not found. "
            f"The should be located in {specifications}."
        )
    return pd.read_csv(specifications, sep="\t")


def create_sessions_tsv_file(
    input_path: Path,
    clinical_data_dir: Path,
    clinical_specifications_folder: Path,
) -> None:
    """Extract the information regarding the sessions and save them in a tsv file.

    Parameters
    ----------
    input_path : Path
        The path to the input folder.

    clinical_data_dir : Path
        The path to the directory to the clinical data files.

    clinical_specifications_folder : Path
        The path to the folder containing the clinical specification files.
    """
    import glob

    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    specifications = _load_specifications(
        clinical_specifications_folder, "sessions.tsv"
    )
    sessions_fields = specifications[StudyName.AIBL.value]
    field_location = specifications[f"{StudyName.AIBL.value} location"]
    sessions_fields_bids = specifications["BIDS CLINICA"]
    fields_dataset = []
    fields_bids = []

    for i in range(0, len(sessions_fields)):
        if not pd.isnull(sessions_fields[i]):
            fields_bids.append(sessions_fields_bids[i])
            fields_dataset.append(sessions_fields[i])

    files_to_read: List[str] = []
    sessions_fields_to_read: List[str] = []
    for i in range(0, len(sessions_fields)):
        # If the i-th field is available
        if not pd.isnull(sessions_fields[i]):
            # Load the file
            file_to_read_path = clinical_data_dir / field_location[i]
            files_to_read.append(glob.glob(str(file_to_read_path))[0])
            sessions_fields_to_read.append(sessions_fields[i])

    rids = pd.read_csv(files_to_read[0], dtype={"text": str}, low_memory=False).RID
    unique_rids = list(set(rids))
    for rid in unique_rids:
        for file in files_to_read:
            df = pd.read_csv(file, dtype={"text": str})
            if len(df.columns) == 1:
                df = pd.read_csv(file, sep=";", low_memory=False)

            visit_code = df.loc[(df["RID"] == rid), "VISCODE"]

            for field in sessions_fields_to_read:
                if field in list(df.columns.values) and field == "MMSCORE":
                    mm_score = df.loc[(df["RID"] == rid), field]
                    mm_score[mm_score == -4] = "n/a"

                elif field in list(df.columns.values) and field == "CDGLOBAL":
                    cd_global = df.loc[(df["RID"] == rid), field]
                    cd_global[cd_global == -4] = "n/a"

                elif field in list(df.columns.values) and field == "DXCURREN":
                    dx_curren = df.loc[(df["RID"] == rid), field]
                    dx_curren[dx_curren == -4] = "n/a"
                    dx_curren[dx_curren == 1] = "CN"
                    dx_curren[dx_curren == 2] = "MCI"
                    dx_curren[dx_curren == 3] = "AD"

                elif field in list(df.columns.values) and field == "EXAMDATE":
                    exam_date = df.loc[(df["RID"] == rid), field]

                elif field in list(df.columns.values) and field == "PTDOB":
                    patient_date_of_birth = df.loc[(df["RID"] == rid), field]

        exam_dates = _clean_exam_dates(
            rid, exam_date.to_list(), visit_code.to_list(), clinical_data_dir
        )
        age = _compute_ages_at_each_exam(patient_date_of_birth.values[0], exam_dates)

        visit_code[visit_code == "bl"] = "M000"
        visit_code = visit_code.str.upper()

        sessions = pd.DataFrame(
            {
                "months": visit_code.str[1:],
                "age": age,
                "MMS": mm_score,
                "cdr_global": cd_global,
                "diagnosis": dx_curren,
                "examination_date": exam_dates,
            }
        )
        sessions = sessions.assign(
            session_id=lambda df: df.months.apply(lambda x: f"ses-M{int(x):03d}")
        )
        cols = sessions.columns.tolist()
        sessions = sessions[cols[-1:] + cols[:-1]]

        bids_id = bids_id_factory(StudyName.AIBL).from_original_study_id(str(rid))

        bids_paths = input_path / bids_id
        if bids_paths.exists():
            sessions.to_csv(
                input_path / bids_id / f"{bids_id}_sessions.tsv",
                sep="\t",
                index=False,
                encoding="utf8",
            )


def _clean_exam_dates(
    rid: str, exam_dates: List[str], visit_codes: List[str], clinical_data_dir: Path
) -> List[str]:
    """Clean the exam dates when necessary by trying to compute them from other sources."""
    from clinica.utils.stream import cprint

    exam_dates_cleaned: List[str] = []
    for visit_code, exam_date in zip(visit_codes, exam_dates):
        if exam_date == "-4":
            exam_date = _find_exam_date_in_other_csv_files(
                rid, visit_code, clinical_data_dir
            ) or _compute_exam_date_from_baseline(visit_code, exam_dates, visit_codes)
            if not exam_date:
                cprint(f"No EXAMDATE for subject %{rid}, at session {visit_code}")
                exam_date = "-4"
        exam_dates_cleaned.append(exam_date)

    return exam_dates_cleaned


def _find_exam_date_in_other_csv_files(
    rid: str, visit_code: str, clinical_data_dir: Path
) -> Optional[str]:
    """Try to find an alternative exam date by searching in other CSV files."""
    for csv_file in _get_cvs_files(clinical_data_dir):
        if "aibl_flutemeta" in csv_file:
            csv_data = pd.read_csv(
                csv_file, low_memory=False, usecols=list(range(0, 36))
            )
        else:
            csv_data = pd.read_csv(csv_file, low_memory=False)
        exam_date = csv_data[(csv_data.RID == rid) & (csv_data.VISCODE == visit_code)]
        if not exam_date.empty and exam_date.iloc[0].EXAMDATE != "-4":
            return exam_date.iloc[0].EXAMDATE
    return None


def _get_cvs_files(clinical_data_dir: Path) -> List[str]:
    """Return a list of paths to CSV files in which an alternative exam date could be found."""
    import glob

    return [
        glob.glob(str(clinical_data_dir / pattern))[0]
        for pattern in (
            "aibl_mri3meta_*.csv",
            "aibl_mrimeta_*.csv",
            "aibl_cdr_*.csv",
            "aibl_flutemeta_*.csv",
            "aibl_mmse_*.csv",
            "aibl_pibmeta_*.csv",
        )
    ]


def _compute_exam_date_from_baseline(
    visit_code: str, exam_dates: List[str], visit_codes: List[str]
) -> Optional[str]:
    """Try to find an alternative exam date by computing the number of months from the visit code."""
    from datetime import datetime

    from dateutil.relativedelta import relativedelta

    baseline_index = visit_codes.index("bl")
    if baseline_index > -1:
        baseline_date = datetime.strptime(exam_dates[baseline_index], "%m/%d/%Y")
        if visit_code != "bl":
            try:
                months = int(visit_code[1:])
            except TypeError:
                raise ValueError(
                    f"Unexpected visit code {visit_code}. Should be in format MXXX."
                    "Ex: M000, M006, M048..."
                )
            exam_date = baseline_date + relativedelta(months=+months)
            return exam_date.strftime("%m/%d/%Y")
    return None


def _compute_ages_at_each_exam(
    patient_date_of_birth: str, exam_dates: List[str]
) -> List[float]:
    """Compute the ages of the patient at each exam date.

    Parameters
    ----------
    patient_date_of_birth : str
        Date of birth of patient ("/%Y" format)

    exam_dates : list of str
        List of exam dates ("%m/%d/%Y" format)

    Return
    ------
    list of float
        The ages of the patient at each exam date.
    """
    from datetime import datetime

    ages: List[float] = []
    date_of_birth = datetime.strptime(patient_date_of_birth, "/%Y")

    for exam_date in exam_dates:
        exam_date = datetime.strptime(exam_date, "%m/%d/%Y")
        delta = exam_date - date_of_birth
        ages.append(round(delta.days / 365.25, 1))

    return ages


def create_scans_tsv_file(
    input_path: Path,
    clinical_data_dir: Path,
    clinical_specifications_folder: Path,
) -> None:
    """Create scans.tsv files for AIBL.

    Parameters
    ----------
    input_path : Path
        The path to the input folder.

    clinical_data_dir : Path
        The path to the directory to the clinical data files.

    clinical_specifications_folder : Path
        The path to the folder containing the clinical specification files.
    """
    import glob
    from os import path

    import clinica.iotools.bids_utils as bids

    specifications = _load_specifications(clinical_specifications_folder, "scans.tsv")
    scans_fields = specifications[bids.StudyName.AIBL.value]
    field_location = specifications[f"{bids.StudyName.AIBL.value} location"]
    scans_fields_bids = specifications["BIDS CLINICA"]
    fields_dataset = []
    fields_bids = []

    # Keep only fields for which there are AIBL fields
    for i in range(0, len(scans_fields)):
        if not pd.isnull(scans_fields[i]):
            fields_bids.append(scans_fields_bids[i])
            fields_dataset.append(scans_fields[i])

    files_to_read = []
    sessions_fields_to_read = []
    for i in range(0, len(scans_fields)):
        # If the i-th field is available
        if not pd.isnull(scans_fields[i]):
            file_to_read_path = clinical_data_dir / field_location[i]
            files_to_read.append(glob.glob(str(file_to_read_path))[0])
            sessions_fields_to_read.append(scans_fields[i])

    bids_ids = [
        path.basename(sub_path) for sub_path in glob.glob(str(input_path / "sub-AIBL*"))
    ]
    ses_dict = {
        bids_id: {"M000": "bl", "M018": "m18", "M036": "m36", "M054": "m54"}
        for bids_id in bids_ids
    }
    scans_dict = bids.create_scans_dict(
        clinical_data_dir,
        bids.StudyName.AIBL,
        clinical_specifications_folder,
        bids_ids,
        "RID",
        "VISCODE",
        ses_dict,
    )
    bids.write_scans_tsv(input_path, bids_ids, scans_dict)



from typing import List, Optional

import click

import clinica.pydra.engine_utils as pydra_utils
from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pydra-t1-volume-register-dartel"


@clinica_pipeline
@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.group_label
# @cli_param.option_group.common_pipelines_options
# @cli_param.option.subjects_sessions_tsv
# @cli_param.option.working_directory
@option.global_option_group
@option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option.dartel_tissues
def cli(
    bids_directory: str,
    caps_directory: str,
    group_label: str,
    dartel_tissues: List[int] = (1, 2, 3),
    n_procs: Optional[int] = None,
) -> None:
    """Inter-subject registration using an existing Dartel template (Pydra engine).

    Parameters
    ----------
    bids_directory : str
        Path to the input BIDS directory.

    caps_directory : str
        Path to the CAPS directory.

    group_label : str
        User-defined identifier to target a specific group of subjects.
        For this pipeline, it is associated to the DARTEL template that was created
        when running the t1-volume pipeline.

    dartel_tissues : List[int], optional
        Indices for selecting the tissues. Possible values are:
            - 1: "graymatter"
            - 2: "whitematter"
            - 3: "csf"
            - 4: "bone"
            - 5: "softtissue"
            - 6: "background"

        Default=[1, 2, 3].

    n_procs : int, optional
        Number of processes to use.

    Notes
    -----
    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_Volume/
    """
    import clinica.pydra.t1_volume.register_dartel.pipeline as pydra_t1vol_rd

    parameters = {"group_label": group_label, "dartel_tissues": dartel_tissues}

    t1_volume_tissue_segmentation_pipeline = pydra_t1vol_rd.t1volume_register_dartel(
        name="t1-volume-tissue-segmentation-pydra",
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters=parameters,
    )
    pydra_utils.run(t1_volume_tissue_segmentation_pipeline, n_procs=n_procs)



from os import PathLike
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd

from clinica.iotools.utils.pipeline_handling import (
    PipelineNameForMetricExtraction,
    pipeline_metric_extractor_factory,
)
from clinica.utils.stream import cprint, log_and_raise, log_and_warn

__all__ = ["create_merge_file"]


def create_merge_file(
    bids_dir: PathLike,
    out_tsv: PathLike,
    caps_dir: Optional[PathLike] = None,
    tsv_file: Optional[PathLike] = None,
    pipelines: Optional[List[str]] = None,
    ignore_scan_files: bool = False,
    ignore_sessions_files: bool = False,
    **kwargs,
):
    """Merge all the TSV files containing clinical data of a BIDS compliant
    dataset and store the result inside a TSV file.

    Parameters
    ----------
    bids_dir : PathLike
        Path to the BIDS folder.

    out_tsv : PathLike
        Path to the output tsv file.

    caps_dir : PathLike, optional
        Path to the CAPS folder.

    tsv_file : PathLike, optional
        Path to a TSV file containing the subjects with their sessions.

    pipelines : list of str, optional
        When adding CAPS information, indicates the pipelines that will be merged.

    ignore_scan_files : bool, optional
        If True the information related to scans is not read.

    ignore_sessions_files : bool, optional
        If True the information related to sessions and scans is not read.
    """
    from os import path

    bids_dir = Path(bids_dir)
    caps_dir = _validate_caps_dir(caps_dir)
    out_path = _validate_output_tsv_path(out_tsv)
    if pipelines:
        pipelines = [
            PipelineNameForMetricExtraction(pipeline) for pipeline in pipelines
        ]
    participants_df, sub_ses_df = _get_participants_and_subjects_sessions_df(
        bids_dir, tsv_file, ignore_sessions_files
    )
    merged_df = _create_merge_file_from_bids(
        bids_dir, sub_ses_df, participants_df, ignore_scan_files, ignore_sessions_files
    )
    merged_df.to_csv(out_path, sep="\t", index=False)
    cprint("End of BIDS information merge.", lvl="debug")
    merged_df.reset_index(drop=True, inplace=True)
    if caps_dir is not None:
        merged_df, merged_summary_df = _add_data_to_merge_file_from_caps(
            caps_dir, merged_df, pipelines, **kwargs
        )
        summary_path = path.splitext(str(out_path))[0] + "_summary.tsv"
        merged_summary_df.to_csv(summary_path, sep="\t", index=False)
        merged_df.to_csv(out_path, sep="\t", index=False)
        cprint("End of CAPS information merge.", lvl="debug")


def _validate_caps_dir(caps_dir: Optional[PathLike] = None) -> Optional[Path]:
    if caps_dir is not None:
        caps_dir = Path(caps_dir)
        if not caps_dir.is_dir():
            log_and_raise(
                f"The path to the CAPS directory ({caps_dir}) is wrong", IOError
            )
    return caps_dir


def _validate_output_tsv_path(out_path: PathLike) -> Path:
    """Validate that provided file path is a TSV file.

    If folders do not exist, this function will create them.
    If provided path is a directory, this function will return
    a file named 'merge.tsv' within this directory.
    """
    out_path = Path(out_path).resolve()
    if out_path.is_dir():
        out_path = out_path / "merge.tsv"
    elif "." not in out_path.name:
        out_path = out_path.with_suffix(".tsv")
    elif out_path.suffix != ".tsv":
        log_and_raise(
            f"Output path extension must be tsv. {out_path} was provided instead.",
            TypeError,
        )
    if out_path.exists():
        log_and_warn(
            f"Path to TSV file {out_path} already exists. The file will be overwritten.",
            UserWarning,
        )
    out_dir = out_path.parent
    if not out_dir.exists():
        out_dir.mkdir(parents=True)
    return out_path


def _get_participants_and_subjects_sessions_df(
    bids_dir: Path,
    tsv_file: Optional[PathLike] = None,
    ignore_sessions_files: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    from clinica.utils.participant import get_subject_session_list

    index_cols = ["participant_id", "session_id"]
    subjects, sessions = get_subject_session_list(
        bids_dir,
        subject_session_file=tsv_file,
        use_session_tsv=(not ignore_sessions_files),
    )
    if (bids_dir / "participants.tsv").is_file():
        participants_df = pd.read_csv(bids_dir / "participants.tsv", sep="\t")
    else:
        participants_df = pd.DataFrame(
            sorted(list(set(subjects))), columns=["participant_id"]
        )
    sub_ses_df = pd.DataFrame(
        [[subject, session] for subject, session in zip(subjects, sessions)],
        columns=index_cols,
    )
    try:
        sub_ses_df.set_index(index_cols, inplace=True, verify_integrity=True)
    except ValueError:
        log_and_warn(
            "Found duplicate subject/session pair. Keeping first occurrence.",
            UserWarning,
        )
        sub_ses_df = sub_ses_df.drop_duplicates(subset=index_cols)
        sub_ses_df.set_index(index_cols, inplace=True)

    return participants_df, sub_ses_df


def _create_merge_file_from_bids(
    bids_dir: Path,
    sub_ses_df: pd.DataFrame,
    participants_df: pd.DataFrame,
    ignore_scan_files: bool = False,
    ignore_sessions_files: bool = False,
) -> pd.DataFrame:
    """Create a merge pandas dataframe for a given BIDS dataset."""
    df = pd.DataFrame(columns=participants_df.columns.values)
    for subject, subject_df in sub_ses_df.groupby(level=0):
        row_subject_data_only = _get_row_data_from_participant_df(
            participants_df, subject
        )
        if ignore_sessions_files:
            rows_session_data_only = (
                _compute_session_rows_for_subject_without_session_files(subject_df)
            )
        else:
            rows_session_data_only = (
                _compute_session_rows_for_subject_with_session_files(
                    bids_dir, subject, subject_df, ignore_scan_files
                )
            )
        for row_session_data_only in rows_session_data_only:
            full_row = pd.concat([row_subject_data_only, row_session_data_only], axis=1)
            df = pd.concat([df, full_row], axis=0)
    return _post_process_merge_file_from_bids(df)


def _get_row_data_from_participant_df(
    participants_df: pd.DataFrame, subject: str
) -> pd.DataFrame:
    """Return, for a given subject, part of a merged dataframe row.
    This part of the full row only contains information found in the participants dataframe.
    In order to make a full row entry, it will be concatenated with information from the sessions and scans.
    """
    row_subject_data = participants_df[participants_df["participant_id"] == subject]
    row_subject_data.reset_index(inplace=True, drop=True)
    if len(row_subject_data) == 0:
        log_and_warn(
            f"Participant {subject} does not exist in participants.tsv",
            UserWarning,
        )
        row_subject_data = pd.DataFrame([[subject]], columns=["participant_id"])

    return row_subject_data


def _compute_session_rows_for_subject_without_session_files(
    subject_df: pd.DataFrame,
) -> List[pd.DataFrame]:
    """Compute all rows with session information for a given subject.
    This will be concatenated with information from the participants TSV file
    to build a full dataframe row with all relevant information.
    This is a default implementation when sessions and scans are ignored.
    """
    return [
        pd.DataFrame([[session]], columns=["session_id"])
        for _, session in subject_df.index.values
    ]


def _compute_session_rows_for_subject_with_session_files(
    bids_dir: Path,
    subject: str,
    subject_df: pd.DataFrame,
    ignore_scan_files: bool,
) -> List[pd.DataFrame]:
    """Compute all rows with session information for a given subject.
    This will be concatenated with information from the participants TSV file
    to build a full dataframe row with all relevant information.
    This function will try to incorporate data from the sessions and scans
    if available.
    """
    from clinica.utils.exceptions import ClinicaDatasetError

    rows = []
    sub_path = bids_dir / subject
    sessions_df = pd.read_csv(sub_path / f"{subject}_sessions.tsv", sep="\t")
    for _, session in subject_df.index.values:
        row_session_df = sessions_df[sessions_df.session_id == session]
        row_session_df.reset_index(inplace=True, drop=True)
        if len(row_session_df) == 0:
            log_and_raise(
                f"The following sessions are not properly formatted : {sessions_df.loc[0, 'session_id']} / {session}",
                ClinicaDatasetError,
            )
        scan_path = bids_dir / subject / session / f"{subject}_{session}_scans.tsv"
        row_scans_df = _get_row_scan_df(scan_path, ignore_scan_files)
        row_df = pd.concat([row_session_df, row_scans_df], axis=1)
        rows.append(row_df.loc[:, ~row_df.columns.duplicated(keep="last")])

    return rows


def _get_row_scan_df(scan_path: Path, ignore_scan_files: bool) -> pd.DataFrame:
    """Return a dataframe for all scans associated with a given subject and session.
    The scan data come from the scan TSV file. If it doesn't exist, an empty dataframe
    is returned.
    """
    row_scans_df = pd.DataFrame()
    if ignore_scan_files or not scan_path.is_file():
        return row_scans_df
    scans_dict = dict()
    scans_df = pd.read_csv(scan_path, sep="\t")
    for idx in scans_df.index.values:
        filepath = scans_df.loc[idx, "filename"]
        if not filepath.endswith(".nii.gz"):
            continue
        filename = Path(filepath).name.split(".")[0]
        modality = "_".join(filename.split("_")[2::])
        for col in scans_df.columns.values:
            if col != "filename":
                value = scans_df.loc[idx, col]
                new_col_name = f"{modality}_{col}"
                scans_dict.update({new_col_name: value})
        json_path = scan_path.parent / f"{filepath.split('.')[0]}.json"
        scans_dict = _add_metadata_from_json(json_path, scans_dict, modality)
    scans_dict = {str(key): str(value) for key, value in scans_dict.items()}
    row_scans_df = pd.DataFrame(scans_dict, index=[0])

    return row_scans_df


def _add_metadata_from_json(json_path: Path, scans_dict: dict, modality: str) -> dict:
    """Add metadata from the provided JSON sidecar file to the scan dictionary."""
    import json

    if json_path.exists():
        try:
            with open(json_path, "r") as f:
                json_dict = json.load(f)
        except json.JSONDecodeError as e:
            log_and_warn(
                f"Couldn't parse the JSON file {json_path}. Ignoring metadata.\nJson error is:\n{e}",
                UserWarning,
            )
            json_dict = {}
        for key, value in json_dict.items():
            new_col_name = f"{modality}_{key}"
            scans_dict.update({new_col_name: value})
    return scans_dict


def _post_process_merge_file_from_bids(merged_df: pd.DataFrame) -> pd.DataFrame:
    """Put participant_id and session_id first, and round numbers."""
    col_list = merged_df.columns.values.tolist()
    col_list.insert(0, col_list.pop(col_list.index("participant_id")))
    col_list.insert(1, col_list.pop(col_list.index("session_id")))
    merged_df = merged_df[col_list]
    tmp = merged_df.select_dtypes(include=[np.number])
    merged_df.loc[:, tmp.columns] = np.round(tmp, 6)

    return merged_df


def _add_data_to_merge_file_from_caps(
    caps_dir: Path,
    merged_df: pd.DataFrame,
    pipelines: Optional[Iterable[PipelineNameForMetricExtraction]] = None,
    **kwargs,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    merged_summary_df = pd.DataFrame()
    if "group_selection" in kwargs and kwargs["group_selection"] is None:
        kwargs.pop("group_selection")
    pipelines = pipelines or PipelineNameForMetricExtraction
    for pipeline in pipelines:
        metric_extractor = pipeline_metric_extractor_factory(pipeline)
        cprint(f"Extracting from CAPS pipeline output: {pipeline.value}...", lvl="info")
        merged_df, summary_df = metric_extractor(caps_dir, merged_df, **kwargs)
        if summary_df is not None and not summary_df.empty:
            merged_summary_df = pd.concat([merged_summary_df, summary_df])
        if summary_df is None or summary_df.empty:
            cprint(
                f"{pipeline.value} outputs were not found in the CAPS folder.",
                lvl="info",
            )

    return _post_process_merged_df_from_caps(merged_df, merged_summary_df)


def _post_process_merged_df_from_caps(
    merged_df: pd.DataFrame,
    merged_summary_df: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    if len(merged_summary_df) == 0:
        log_and_raise(
            (
                "No outputs were found for any pipeline in the CAPS folder. "
                "The output only contains BIDS information."
            ),
            FileNotFoundError,
        )
    columns = merged_df.columns.values.tolist()
    merged_summary_df.reset_index(inplace=True, drop=True)
    for idx in merged_summary_df.index:
        first_column_name = merged_summary_df.loc[idx, "first_column_name"]
        last_column_name = merged_summary_df.loc[idx, "last_column_name"]
        merged_summary_df.loc[idx, "first_column_index"] = columns.index(
            first_column_name
        )
        merged_summary_df.loc[idx, "last_column_index"] = columns.index(
            last_column_name
        )
    tmp = merged_df.select_dtypes(include=[np.number])
    merged_df.loc[:, tmp.columns] = np.round(tmp, 12)

    return merged_df, merged_summary_df



"""This module contains utilities for PET data handling."""
import os
import typing as ty
from enum import Enum
from pathlib import Path

import pandas as pd


class Tracer(str, Enum):
    """BIDS label for PET tracers.

    Follows the convention proposed in the PET section of the BIDS specification.

    See: https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/09-positron-emission-tomography.html
    """

    PIB = "11CPIB"
    AV1451 = "18FAV1451"
    AV45 = "18FAV45"
    FBB = "18FFBB"
    FDG = "18FFDG"
    FMM = "18FFMM"


class SUVRReferenceRegion(str, Enum):
    PONS = "pons"
    CEREBELLUM_PONS = "cerebellumPons"
    PONS2 = "pons2"
    CEREBELLUM_PONS2 = "cerebellumPons2"


class ReconstructionMethod(str, Enum):
    """BIDS label for PET reconstruction methods.

    Follows the convention proposed in the PET section of the BIDS specification.

    See: https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/09-positron-emission-tomography.html#pet-recording-data

    For ADNI specific reconstruction methods, see:

    https://adni.loni.usc.edu/data-samples/adni-data/neuroimaging/pet/
    """

    # Reconstruction methods defined in the BIDS specifications
    STATIC = "nacstat"
    DYNAMIC = "nacdyn"
    STATIC_ATTENUATION_CORRECTION = "acstat"
    DYNAMIC_ATTENUATION_CORRECTION = "acdyn"

    # ADNI specific reconstruction methods
    CO_REGISTERED_DYNAMIC = "coregdyn"  # Corresponds to ADNI processing steps 1
    CO_REGISTERED_AVERAGED = "coregavg"  # Corresponds to ADNI processing steps 2
    CO_REGISTERED_STANDARDIZED = "coregstd"  # Corresponds to ADNI processing steps 3
    COREGISTERED_ISOTROPIC = "coregiso"  # Corresponds to ADNI processing steps 4



from pathlib import Path
from typing import Optional, Union

from rich import box
from rich.console import Console
from rich.table import Table

from clinica.iotools.bids_dataset_description import BIDSDatasetDescription
from clinica.utils.caps import CAPSDatasetDescription
from clinica.utils.inputs import DatasetType

__all__ = ["describe"]


def describe(dataset: Union[str, Path]) -> None:
    """Describe a dataset in BIDS or CAPS format.

    Parameters
    ----------
    dataset : str or Path
        The path to the BIDS or CAPS dataset to describe.
    """
    description = (
        CAPSDatasetDescription if _is_caps(dataset) else BIDSDatasetDescription
    ).from_file(Path(dataset) / "dataset_description.json")
    console = Console()
    console.print(_build_dataset_table(description))
    if (processing_table := _build_processing_table(description)) is not None:
        console.print(processing_table)


def _is_caps(dataset: Union[str, Path]) -> bool:
    import json

    from clinica.utils.inputs import DatasetType

    dataset_description = Path(dataset) / "dataset_description.json"
    with open(dataset_description, "r") as fp:
        metadata = json.load(fp)

    return metadata["DatasetType"] == DatasetType.DERIVATIVE


def _build_dataset_table(
    description: Union[BIDSDatasetDescription, CAPSDatasetDescription],
) -> Table:
    dataset_table = Table(title="Dataset information", box=box.ROUNDED)
    for column, color in zip(
        ["Name", "Type", "BIDS Specifications Version"],
        ["bright_cyan", "bright_yellow", "bright_magenta"],
    ):
        dataset_table.add_column(column, justify="right", style=color)
    row = (description.name, description.dataset_type, str(description.bids_version))
    if description.dataset_type == DatasetType.DERIVATIVE:
        dataset_table.add_column(
            "CAPS Specifications Version", justify="right", style="bright_green"
        )
        row = (
            description.name,
            description.dataset_type,
            str(description.bids_version),
            str(description.caps_version),
        )
    dataset_table.add_row(*row)

    return dataset_table


def _build_processing_table(
    description: Union[BIDSDatasetDescription, CAPSDatasetDescription],
) -> Optional[Table]:
    if description.dataset_type == DatasetType.RAW:
        return None
    processing_table = Table(title="Processing information", box=box.ROUNDED)
    for column, color in zip(
        [
            "Name",
            "Date",
            "Author",
            "Machine",
            "InputPath",
            "ClinicaVersion",
            "Dependencies",
        ],
        [
            "bright_cyan",
            "bright_yellow",
            "bright_magenta",
            "bright_green",
            "bright_red",
            "bright_blue",
            "bright_green",
        ],
    ):
        processing_table.add_column(column, justify="right", style=color)
    for processing in description.processing:
        dependency_table = Table(title="Dependencies", box=box.ROUNDED)
        for column in ("Name", "VersionConstraint", "InstalledVersion"):
            dependency_table.add_column(column, justify="right")
        for dependency in processing.dependencies:
            dependency_table.add_row(
                dependency.name,
                str(dependency.version_constraint),
                str(dependency.installed_version),
            )
        processing_table.add_row(
            processing.name,
            str(processing.date),
            processing.author,
            processing.machine,
            processing.input_path,
            str(processing.clinica_version),
            dependency_table,
        )
    return processing_table



import abc
import os.path as path

import numpy as np
from pandas.io import parsers

import clinica.pipelines.machine_learning.ml_utils as utils
import clinica.pipelines.machine_learning.region_based_io as rbio
import clinica.pipelines.machine_learning.tsv_based_io as tbio
import clinica.pipelines.machine_learning.vertex_based_io as vtxbio
import clinica.pipelines.machine_learning.voxel_based_io as vbio
from clinica.pipelines.machine_learning import base
from clinica.utils.stream import cprint


class CAPSInput(base.MLInput):
    def __init__(self, input_params):
        super().__init__(input_params)

        self._images = None

        subjects_visits = parsers.read_csv(
            self._input_params["subjects_visits_tsv"], sep="\t"
        )
        if list(subjects_visits.columns.values) != ["participant_id", "session_id"]:
            raise Exception("Subjects and visits file is not in the correct format.")
        self._subjects = list(subjects_visits.participant_id)
        self._sessions = list(subjects_visits.session_id)

        diagnoses = parsers.read_csv(self._input_params["diagnoses_tsv"], sep="\t")
        if "diagnosis" not in list(diagnoses.columns.values):
            raise Exception("Diagnoses file is not in the correct format.")
        self._diagnoses = list(diagnoses.diagnosis)

        if self._input_params["image_type"] not in ["T1w", "PET"]:
            raise ValueError(
                f"Incorrect image type. It must be one of the values 'T1w', 'PET' "
                f"(given value: {self._input_params['image_type']})"
            )

        if self._input_params["precomputed_kernel"] is not None:
            if isinstance(self._input_params["precomputed_kernel"], np.ndarray):
                if self._input_params["precomputed_kernel"].shape == (
                    len(self._subjects),
                    len(self._subjects),
                ):
                    self._kernel = self._input_params["precomputed_kernel"]
                else:
                    raise Exception(
                        """Precomputed kernel provided is not in the correct format.
                    It must be a numpy.ndarray object with number of rows and columns equal to the number of subjects,
                    or a filename to a numpy txt file containing an object with the described format."""
                    )
            elif type(self._input_params["precomputed_kernel"] == str):
                self._kernel = np.loadtxt(self._input_params["precomputed_kernel"])
            else:
                raise Exception(
                    """Precomputed kernel provided is not in the correct format.
                It must be a numpy.ndarray object with number of rows and columns equal to the number of subjects,
                or a filename to a numpy txt file containing an object with the described format."""
                )

    @abc.abstractmethod
    def get_images(self):
        """
        Returns: a list of filenames
        """

    @abc.abstractmethod
    def get_x(self):
        """
        Returns: a numpy 2d-array.
        """

    def get_y(self):
        """
        Returns: a list of integers. Each integer represents a class.
        """
        if self._y is not None:
            return self._y

        unique = sorted(list(set(self._diagnoses)))
        self._y = np.array([unique.index(x) for x in self._diagnoses])
        return self._y

    def get_kernel(
        self, kernel_function=utils.gram_matrix_linear, recompute_if_exists=False
    ):
        """
        Returns: a numpy 2d-array.
        """
        if self._kernel is not None and not recompute_if_exists:
            return self._kernel

        if self._x is None:
            self.get_x()

        cprint("Computing kernel ...")
        self._kernel = kernel_function(self._x)
        cprint("Kernel computed")
        return self._kernel

    def save_kernel(self, output_dir):
        """

        Args:
            output_dir:

        Returns:

        """
        if self._kernel is not None:
            filename = path.join(output_dir, "kernel.txt")
            np.savetxt(filename, self._kernel)
            return filename
        raise Exception(
            "Unable to save the kernel. Kernel must have been computed before."
        )

    @abc.abstractmethod
    def save_weights_as_nifti(self, weights, output_dir):
        pass

    @staticmethod
    def get_default_parameters():
        parameters_dict = {}
        parameters_dict.setdefault("caps_directory", None)
        parameters_dict.setdefault("subjects_visits_tsv", None)
        parameters_dict.setdefault("diagnoses_tsv", None)
        parameters_dict.setdefault("group_label", None)
        parameters_dict.setdefault("image_type", None)
        parameters_dict.setdefault("precomputed_kernel", None)

        return parameters_dict


class CAPSVoxelBasedInput(CAPSInput):
    def __init__(self, input_params):
        super().__init__(input_params)

        self._orig_shape = None
        self._data_mask = None

        if self._input_params["modulated"] not in ["on", "off"]:
            raise Exception(
                "Incorrect modulation parameter. It must be one of the values 'on' or 'off'"
            )

    def get_images(self):
        """
        Returns: a list of filenames
        """
        from clinica.utils.input_files import pet_volume_normalized_suvr_pet
        from clinica.utils.inputs import clinica_file_reader

        if self._images is not None:
            return self._images

        if self._input_params["image_type"] == "T1w":
            if self._input_params["fwhm"] == 0:
                fwhm_key_value = ""
            else:
                fwhm_key_value = f"_fwhm-{self._input_params['fwhm']}mm"

            self._images = [
                path.join(
                    self._input_params["caps_directory"],
                    "subjects",
                    self._subjects[i],
                    self._sessions[i],
                    "t1",
                    "spm",
                    "dartel",
                    f"group-{self._input_params['group_label']}",
                    f"{self._subjects[i]}_{self._sessions[i]}_T1w"
                    f"_segm-graymatter_space-Ixi549Space_modulated-{self._input_params['modulated']}{fwhm_key_value}_probability.nii.gz",
                )
                for i in range(len(self._subjects))
            ]

            for image in self._images:
                if not path.exists(image):
                    raise Exception("File %s doesn't exists." % image)

        elif self._input_params["image_type"] == "PET":
            caps_files_information = pet_volume_normalized_suvr_pet(
                acq_label=self._input_params["acq_label"],
                group_label=self._input_params["group_label"],
                suvr_reference_region=self._input_params["suvr_reference_region"],
                use_brainmasked_image=True,
                use_pvc_data=self._input_params["use_pvc_data"],
                fwhm=self._input_params["fwhm"],
            )
            self._images, _ = clinica_file_reader(
                self._subjects,
                self._sessions,
                self._input_params["caps_directory"],
                caps_files_information,
            )
        else:
            raise ValueError(
                f"Unknown image type (given value: {self._input_params['image_type']})"
            )

        return self._images

    def get_x(self):
        """
        Returns: a numpy 2d-array.
        """
        if self._x is not None:
            return self._x

        cprint(f"Loading {len(self.get_images())} subjects")
        self._x, self._orig_shape, self._data_mask = vbio.load_data(
            self._images, mask=self._input_params["mask_zeros"]
        )
        cprint("Subjects loaded")

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):
        if self._images is None:
            self.get_images()

        output_filename = path.join(output_dir, "weights.nii.gz")
        data = vbio.revert_mask(weights, self._data_mask, self._orig_shape)
        vbio.weights_to_nifti(data, self._images[0], output_filename)

    @staticmethod
    def get_default_parameters():
        parameters_dict = super(
            CAPSVoxelBasedInput, CAPSVoxelBasedInput
        ).get_default_parameters()
        # t1-volume / pet-volume
        parameters_dict.setdefault("fwhm", 0)
        # t1-volume / pet-volume ?
        parameters_dict.setdefault("mask_zeros", True)
        # t1-volume
        parameters_dict.setdefault("modulated", "on")
        # pet-volume
        parameters_dict.setdefault("acq_label", None)
        parameters_dict.setdefault("suvr_reference_region", None)
        parameters_dict.setdefault("use_pvc_data", False)

        return parameters_dict


class CAPSRegionBasedInput(CAPSInput):
    def __init__(self, input_params):
        from clinica.utils.atlas import AtlasName

        super().__init__(input_params)
        AtlasName(self._input_params["atlas"])

    def get_images(self):
        """

        Returns: a list of filenames

        """
        if self._images is not None:
            return self._images

        if self._input_params["image_type"] == "T1w":
            self._images = [
                path.join(
                    self._input_params["caps_directory"],
                    "subjects",
                    self._subjects[i],
                    self._sessions[i],
                    "t1",
                    "spm",
                    "dartel",
                    f"group-{self._input_params['group_label']}",
                    "atlas_statistics",
                    f"{self._subjects[i]}_{self._sessions[i]}_T1w"
                    f"_space-{self._input_params['atlas']}_map-graymatter_statistics.tsv",
                )
                for i in range(len(self._subjects))
            ]

        elif self._input_params["image_type"] == "PET":
            if self._input_params["use_pvc_data"]:
                pvc_key_value = "_pvc-rbv"
            else:
                pvc_key_value = ""

            self._images = [
                path.join(
                    self._input_params["caps_directory"],
                    "subjects",
                    self._subjects[i],
                    self._sessions[i],
                    "pet",
                    "preprocessing",
                    f"group-{self._input_params['group_label']}",
                    "atlas_statistics",
                    f"{self._subjects[i]}_{self._sessions[i]}_trc-{self._input_params['acq_label'].value}_pet"
                    f"_space-{self._input_params['atlas']}{pvc_key_value}"
                    f"_suvr-{self._input_params['suvr_reference_region'].value}_statistics.tsv",
                )
                for i in range(len(self._subjects))
            ]
        else:
            raise ValueError(
                f"Unknown image type (given value: {self._input_params['image_type']})"
            )

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images

    def get_x(self):
        """
        Returns: a numpy 2d-array.
        """
        if self._x is not None:
            return self._x

        cprint(f"Loading {len(self.get_images())} subjects")
        self._x = rbio.load_data(self._images, self._subjects)
        cprint("Subjects loaded")

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):
        """

        Args:
            weights:
            output_dir:

        Returns:

        """
        output_filename = path.join(output_dir, "weights.nii.gz")
        rbio.weights_to_nifti(weights, self._input_params["atlas"], output_filename)

    @staticmethod
    def get_default_parameters():
        parameters_dict = super(
            CAPSRegionBasedInput, CAPSRegionBasedInput
        ).get_default_parameters()

        # t1-volume / pet-volume
        parameters_dict.setdefault("atlas", None)
        # t1-volume / pet-volume ?
        parameters_dict.setdefault("mask_zeros", True)
        # pet-volume
        parameters_dict.setdefault("acq_label", None)
        parameters_dict.setdefault("suvr_reference_region", None)
        parameters_dict.setdefault("use_pvc_data", False)

        return parameters_dict


class CAPSVertexBasedInput(CAPSInput):
    def __init__(self, input_params):
        super().__init__(input_params)

    def get_images(self):
        """
        returns list of filenames
        """
        import os

        if self._images is not None:
            return self._images

        if self._input_params["image_type"] == "PET":
            self._images = []
            hemi = ["lh", "rh"]
            for i in range(len(self._subjects)):
                self._images.append(
                    [
                        os.path.join(
                            self._input_params["caps_directory"],
                            "subjects",
                            self._subjects[i],
                            self._sessions[i],
                            "pet",
                            "surface",
                            f"{self._subjects[i]}_{self._sessions[i]}_trc-{self._input_params['acq_label'].value}_pet"
                            f"_space-fsaverage_suvr-{self._input_params['suvr_reference_region'].value}"
                            f"_pvc-iy_hemi-{h}_fwhm-{self._input_params['fwhm']}_projection.mgh",
                        )
                        for h in hemi
                    ]
                )
            missing_files = []
            missing_files_string_error = ""
            for img in self._images:
                for side in img:
                    if not os.path.exists(side):
                        missing_files.append(side)
                        missing_files_string_error += side + "\n"
            if len(missing_files) > 0:
                raise IOError(
                    f"Could not find the following files:\n"
                    f"{missing_files_string_error}\n"
                    f"{len(missing_files)} files missing."
                )
        else:
            raise ValueError(
                f"Unknown image type (given value: {self._input_params['image_type']})"
            )
        return self._images

    def get_x(self):
        """
        Returns numpy 2D array
        """
        if self._x is not None:
            return self._x

        cprint(f"Loading  str({len(self.get_images())} subjects")
        self._x = vtxbio.load_data(self._images)
        cprint(f"{len(self._x)} subjects loaded")
        return self._x

    def save_weights_as_datasurface(self, weights, output_dir):
        import os

        import nibabel as nib
        import numpy as np

        if self._images is None:
            self.get_images()

        sample = nib.load(self._images[0][0])

        infinite_norm = np.max(np.abs(weights))

        left_hemi_data = np.atleast_3d(
            np.divide(weights[: np.int(weights.size / 2)], infinite_norm)
        )
        left_hemi_mgh = nib.MGHImage(
            left_hemi_data, affine=sample.affine, header=sample.header
        )
        nib.save(left_hemi_mgh, os.path.join(output_dir, "weights_lh.mgh"))

        right_hemi_data = np.atleast_3d(
            np.divide(weights[np.int(weights.size / 2) :], infinite_norm)
        )
        right_hemi_mgh = nib.MGHImage(
            right_hemi_data, affine=sample.affine, header=sample.header
        )
        nib.save(right_hemi_mgh, os.path.join(output_dir, "weights_rh.mgh"))

    def save_weights_as_nifti(self, weights, output_dir):
        pass

    @staticmethod
    def get_default_parameters():
        parameters_dict = super(
            CAPSVertexBasedInput, CAPSVertexBasedInput
        ).get_default_parameters()

        # pet-surface
        parameters_dict.setdefault("fwhm", 0)
        parameters_dict.setdefault("acq_label", None)
        parameters_dict.setdefault("suvr_reference_region", None)

        return parameters_dict


class CAPSTSVBasedInput(CAPSInput):
    def __init__(self, input_params):
        from clinica.utils.atlas import VOLUME_ATLASES

        super().__init__(input_params)

        if self._input_params["atlas"] not in VOLUME_ATLASES:
            raise ValueError(
                f"Incorrect atlas name (given value: {self._input_params['atlas']}). "
                f"It must be one of {VOLUME_ATLASES}"
            )

    def get_images(self):
        """
        Returns: string
        """

    def get_x(self):
        """
        Returns: a numpy 2d-array.
        """

        # if self._x is not None:
        #    return self._x

        cprint("Loading TSV subjects")

        self._x = tbio.load_data(
            f"group-{self._input_params['group_label']}_T1w_space-{self._input_params['atlas']}_map-graymatter",
            self._input_params["caps_directory"],
            self._subjects,
            self._sessions,
            self._input_params["dataset"],
        )

        cprint("Subjects loaded")

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):
        """

        Args:
            weights:
            output_dir:

        Returns:

        """

        # output_filename = path.join(output_dir, 'weights.nii.gz')

        # rbio.weights_to_nifti(weights, self._input_params['atlas'], output_filename)

    @staticmethod
    def get_default_parameters():
        parameters_dict = super(
            CAPSTSVBasedInput, CAPSTSVBasedInput
        ).get_default_parameters()

        # ???
        parameters_dict.setdefault("dataset", None)
        # pet-volume
        parameters_dict.setdefault("acq_label", None)
        parameters_dict.setdefault("suvr_reference_region", None)
        parameters_dict.setdefault("use_pvc_data", False)

        return parameters_dict


class CAPSVoxelBasedInputREGSVM(CAPSVoxelBasedInput):
    def get_images(self):
        """
        Returns: a list of filenames
        """
        if self._images is not None:
            return self._images

        if self._input_params["image_type"] == "T1w":
            if self._input_params["fwhm"] == 0:
                fwhm_key_value = ""
            else:
                fwhm_key_value = f"_fwhm-{self._input_params['fwhm']}mm"

            self._images = [
                path.join(
                    self._input_params["caps_directory"],
                    f"regul_{self._subjects[i]}_{self._sessions[i]}_T1w"
                    f"_segm-graymatter_space-Ixi549Space_modulated-{self._input_params['modulated']}{fwhm_key_value}_probability.nii",
                )
                for i in range(len(self._subjects))
            ]
        elif self._input_params["image_type"] == "PET":
            if self._input_params["fwhm"] == 0:
                fwhm_key_value = ""
            else:
                fwhm_key_value = f"_fwhm-{self._input_params['fwhm']}mm"

            if self._input_params["use_pvc_data"]:
                pvc_key_value = "_pvc-rbv"
            else:
                pvc_key_value = ""

            self._images = [
                path.join(
                    self._input_params["caps_directory"],
                    "subjects",
                    self._subjects[i],
                    self._sessions[i],
                    "pet",
                    "preprocessing",
                    f"group-{self._input_params['group_label']}",
                    f"{self._subjects[i]}_{self._sessions[i]}_trc-{self._input_params['acq_label'].value}_pet"
                    f"_space-Ixi549Space{pvc_key_value}_suvr-{self._input_params['suvr_reference_region'].value}"
                    f"_mask-brain{fwhm_key_value}_pet.nii.gz",
                )
                for i in range(len(self._subjects))
            ]
        else:
            raise ValueError(
                f"Unknown image type (given value: {self._input_params['image_type']})"
            )

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images


class TsvInput(base.MLInput):
    def __init__(self, input_params):
        super().__init__(input_params)

        import pandas as pd

        self._dataframe = pd.read_csv(input_params["data_tsv"], sep="\t")

        if not input_params["columns"]:
            raise Exception("List of columns to use as input can not be empty.")

    def get_x(self):
        self._x = self._dataframe.as_matrix(self._input_params["columns"])
        return self._x

    def get_y(self):
        unique = list(set(self._dataframe["diagnosis"]))
        self._y = np.array([unique.index(x) for x in self._dataframe["diagnosis"]])
        return self._y

    def get_kernel(
        self, kernel_function=utils.gram_matrix_linear, recompute_if_exists=False
    ):
        """
        Returns: a numpy 2d-array.
        """

        if self._kernel is not None and not recompute_if_exists:
            return self._kernel

        if self._x is None:
            self.get_x()

        cprint("Computing kernel ...")
        self._kernel = kernel_function(self._x)
        cprint("Kernel computed")
        return self._kernel

    @staticmethod
    def get_default_parameters():
        parameters_dict = {}
        parameters_dict.setdefault("data_tsv", None)
        parameters_dict.setdefault("columns", None)

        return parameters_dict



"""Module for converting DWI of ADNI."""

from pathlib import Path
from typing import Any, Iterable, Optional

import pandas as pd

__all__ = ["convert_dwi"]


def convert_dwi(
    source_dir: Path,
    csv_dir: Path,
    destination_dir: Path,
    conversion_dir: Path,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
):
    """Convert DW images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    destination_dir : Path
        The path to the destination BIDS directory.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        ADNIModalityConverter,
        load_clinical_csv,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    cprint(
        f"Calculating paths of DWI images. Output will be stored in {conversion_dir}.",
        lvl="info",
    )
    images = _compute_dwi_paths(source_dir, csv_dir, subjects, conversion_dir)
    cprint("Paths of DWI images found. Exporting images into BIDS ...", lvl="info")
    paths_to_bids(
        images,
        destination_dir,
        ADNIModalityConverter.DWI,
        mod_to_update=mod_to_update,
        n_procs=n_procs,
    )
    cprint(msg="DWI conversion done.", lvl="debug")


def _compute_dwi_paths(
    source_dir: Path,
    csv_dir: Path,
    subjects: Iterable[str],
    conversion_dir: Path,
) -> pd.DataFrame:
    """Compute paths to DW images to convert to BIDS.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    subjects : list of str
        The subjects list.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    Returns
    -------
    images : pd.DataFrame
        A pandas dataframe that contains the path to all the DWI images to convert.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv

    from ._image_path_utils import find_image_path
    from ._visits_utils import visits_to_timepoints

    dwi_df = _initialize_dwi_df()
    dwi_dfs_list = []
    # Loading needed .csv files
    adni_merge = load_clinical_csv(csv_dir, "ADNIMERGE")
    mayo_mri_qc = load_clinical_csv(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15")
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "DTI"]
    mri_list = load_clinical_csv(csv_dir, "MRILIST")

    # Selecting only DTI images that are not Multiband, processed or enhanced images
    mri_list = mri_list[mri_list.SEQUENCE.str.contains("dti", case=False, na=False)]
    mri_list = mri_list[
        mri_list.SEQUENCE.map(
            lambda x: not any(
                subs in x for subs in ("MB", "ADC", "FA", "TRACEW", "Enhanced", "Reg")
            )
        )
    ]
    for subject in subjects:
        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and
        # sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subject]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")

        mri_list_subj = mri_list[mri_list.SUBJECT == subject]
        mri_list_subj = mri_list_subj.sort_values("SCANDATE")

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subject[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(subject, mri_list_subj, adnimerge_subj, "DWI")

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]
            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            axial = _select_dwi_image(
                subject, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj
            )
            if axial is not None:
                row_to_append = pd.DataFrame(axial, index=["i"])
                dwi_dfs_list.append(row_to_append)
    if dwi_dfs_list:
        dwi_df = pd.concat(dwi_dfs_list, ignore_index=True)

    # Removing known exceptions from images to convert
    if not dwi_df.empty:
        error_ind = dwi_df.index[
            dwi_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in _get_known_conversion_errors()),
                axis=1,
            )
        ]
        dwi_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(dwi_df, source_dir, "DWI")
    images.to_csv(conversion_dir / "dwi_paths.tsv", sep="\t", index=False)

    return images


def _initialize_dwi_df() -> pd.DataFrame:
    return pd.DataFrame(
        columns=[
            "Subject_ID",
            "VISCODE",
            "Visit",
            "Sequence",
            "Scan_Date",
            "Study_ID",
            "Series_ID",
            "Image_ID",
            "Field_Strength",
        ]
    )


def _get_known_conversion_errors() -> Iterable[tuple[str, str]]:
    return [
        ("029_S_2395", "m60"),
        ("029_S_0824", "m108"),
        ("029_S_0914", "m108"),
        ("027_S_2219", "m36"),
        ("129_S_2332", "m12"),
        ("029_S_4384", "m48"),
        ("029_S_4385", "m48"),
        ("029_S_4585", "m48"),
        ("016_S_4591", "m24"),
        ("094_S_4630", "m06"),
        ("094_S_4649", "m06"),
        ("029_S_5219", "m24"),
        ("094_S_2238", "m48"),
        ("129_S_4287", "bl"),
        ("007_S_4611", "m03"),
        ("016_S_4638", "bl"),
        ("027_S_5118", "bl"),
        ("098_S_4018", "bl"),
        ("098_S_4003", "m12"),
        ("016_S_4584", "m24"),
        ("016_S_5007", "m12"),
        ("129_S_2347", "m06"),
        ("129_S_4220", "bl"),
        ("007_S_2058", "m12"),
        ("016_S_2007", "m06"),
        ("020_S_6358", "bl"),
        ("114_S_6039", "m12"),
        ("114_S_6057", "bl"),
        ("153_S_6274", "bl"),
        ("006_S_4485", "m84"),
        ("153_S_6237", "bl"),
        ("153_S_6336", "bl"),
        ("153_S_6450", "bl"),
        ("003_S_4441", "m12"),
        # Eq_1 Images
        ("006_S_6252", "m12"),
        ("010_S_0419", "m156"),
        ("127_S_0259", "m156"),
        ("129_S_6830", "bl"),
        ("135_S_6104", "m24"),
        ("021_S_5237", "m84"),
        ("027_S_2245", "m120"),
        ("099_S_6396", "m24"),
        ("126_S_4507", "m96"),
        ("126_S_4891", "m96"),
        ("126_S_4896", "m96"),
        ("126_S_6559", "m24"),
        ("126_S_6724", "m12"),
        ("127_S_6203", "m24"),
        ("127_S_6330", "m24"),
        ("127_S_6512", "m24"),
        ("127_S_6549", "m24"),
        ("129_S_6459", "m24"),
        ("129_S_6763", "m12"),
        ("129_S_6784", "m12"),
        ("129_S_6830", "m12"),
        ("135_S_6446", "m24"),
        ("301_S_6224", "m36"),
        # Missing .bval and .bvec
        ("020_S_5203", "m72"),
        ("020_S_6185", "m24"),
        ("020_S_6513", "m12"),
        ("021_S_0178", "m156"),
        ("153_S_6336", "m12"),
        ("153_S_6755", "bl"),
        ("020_S_6227", "m24"),
        ("020_S_6449", "m24"),
        ("020_S_6513", "m24"),
        # Volume mismatch between .nii and .bvec / .bval
        ("006_S_6610", "bl"),
        ("006_S_6682", "bl"),
        ("006_S_6696", "bl"),
        ("006_S_6770", "bl"),
        ("029_S_6289", "bl"),
        ("130_S_6043", "bl"),
        ("130_S_6329", "bl"),
        ("027_S_6183", "m24"),
        ("123_S_6891", "bl"),
        # Several output images
        ("029_S_2395", "m72"),
    ]


def _select_dwi_image(
    subject_id: str,
    timepoint,
    visit_str: str,
    visit_mri_list: list,
    mri_qc_subj: pd.DataFrame,
) -> Optional[dict[str, Any]]:
    """Select a DWI image based on QC information.

    One image among those in the input list is chosen according to QC
    and then corresponding metadata is extracted to a dictionary.

    Parameters
    ----------
    subject_id : str
        The subject identifier.

    timepoint : ??
        The visit code.

    visit_str : str
        The visit name.

    visit_mri_list : list
        The list of images metadata.

    mri_qc_subj : pd.Dataframe
        DataFrame containing list of QC of scans for the subject.

    Returns
    -------
    dict or None:
        Contains image metadata.
        Returns None if no image was found.
    """
    from clinica.iotools.converter_utils import replace_sequence_chars

    from ._qc_utils import select_image_qc

    selected_image_id = select_image_qc(
        list(visit_mri_list.IMAGEUID),
        mri_qc_subj,
    )
    if selected_image_id is None:
        return None
    selected_scan = visit_mri_list[visit_mri_list.IMAGEUID == selected_image_id].iloc[0]

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": replace_sequence_chars(selected_scan.SEQUENCE),
        "Scan_Date": selected_scan["SCANDATE"],
        "Study_ID": str(int(selected_scan.STUDYID)),
        "Series_ID": str(int(selected_scan.SERIESID)),
        "Image_ID": str(int(selected_scan.IMAGEUID)),
        "Field_Strength": selected_scan.MAGSTRENGTH,
    }



"""This module contains utilities to check dependencies before running Clinica.

These functions can check binaries, software (e.g. FreeSurfer) or toolboxes (e.g. SPM).
"""
import functools
import json
import os
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Optional, Tuple, Union

from attrs import define
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from clinica.utils.exceptions import ClinicaMissingDependencyError
from clinica.utils.stream import (
    LoggingLevel,
    cprint,
    get_logging_level,
    log_and_raise,
    log_and_warn,
)

__all__ = [
    "ThirdPartySoftware",
    "SoftwareDependency",
    "SoftwareEnvironmentVariable",
    "get_fsl_home",
    "get_freesurfer_home",
    "get_mcr_home",
    "get_spm_standalone_home",
    "get_spm_home",
    "is_binary_present",
    "check_binary",
    "check_environment_variable",
    "check_software",
    "get_software_min_version_supported",
    "get_software_version",
]


class ThirdPartySoftware(str, Enum):
    """Possible third party software that clinica depends on."""

    ANTS = "ants"
    CONVERT3D = "convert3d"
    DCM2NIIX = "dcm2niix"
    FREESURFER = "freesurfer"
    FSL = "fsl"
    MATLAB = "matlab"
    MCR = "MCR"
    MRTRIX = "mrtrix"
    PETPVC = "petpvc"
    SPM = "spm"
    SPMSTANDALONE = "spm standalone"


@define
class SoftwareDependency:
    """Class modeling a dependency on a third-party software.

    Attributes
    ----------
    name : ThirdPartySoftware
        The name of the software that is a dependency.

    version_constraint : SpecifierSet
        The constraint on the version of the software that Clinica imposes.

    installed_version : Version, optional
        The version of the software installed on the machine.
        If the software is not installed, this value is set to None.
    """

    name: ThirdPartySoftware
    version_constraint: SpecifierSet
    installed_version: Optional[Version]

    @classmethod
    def from_strings(cls, name: str, version_constraint: str):
        try:
            installed_version = get_software_version(name)
        except Exception:
            installed_version = None
        if version_constraint == "":
            version_constraint = ">=0.0.0"
        return cls(
            ThirdPartySoftware(name),
            SpecifierSet(version_constraint),
            installed_version,
        )

    @classmethod
    def from_dict(cls, d: dict):
        return cls.from_strings(d["name"], d["version"])

    def to_dict(self) -> dict:
        return {
            "name": self.name.value,
            "version_constraint": str(self.version_constraint),
            "installed_version": str(self.installed_version)
            if self.installed_version
            else "",
        }

    def __str__(self) -> str:
        return json.dumps(self.to_dict())

    def __repr__(self) -> str:
        return self.__str__()

    def check(self):
        check_software(self.name, log_level="error", specifier=self.version_constraint)

    def is_satisfied(self) -> bool:
        try:
            self.check()
        except ClinicaMissingDependencyError:
            return False
        return True


class SoftwareEnvironmentVariable:
    """Represents an environment variable name linked to a specific software.

    Attributes
    ----------
    name : str
        The name of the environment variable.

    software : ThirdPartySoftware
        The name of the software related to the environment variable.
    """

    def __init__(self, name: str, software: Union[str, ThirdPartySoftware]):
        self.name = name
        self.software = ThirdPartySoftware(software)


def get_fsl_home() -> Path:
    """Return the path to the home directory of FSL."""
    return check_environment_variable(
        SoftwareEnvironmentVariable("FSLDIR", ThirdPartySoftware.FSL)
    )


def get_freesurfer_home() -> Path:
    """Return the path to the home directory of Freesurfer."""
    return check_environment_variable(
        SoftwareEnvironmentVariable("FREESURFER_HOME", ThirdPartySoftware.FREESURFER)
    )


def get_mcr_home() -> Path:
    """Return the path to the home directory of Matlab MCR."""
    return check_environment_variable(
        SoftwareEnvironmentVariable("MCR_HOME", ThirdPartySoftware.MCR)
    )


def get_spm_standalone_home() -> Path:
    """Return the path to the home directory of SPM standalone"""
    return check_environment_variable(
        SoftwareEnvironmentVariable(
            "SPMSTANDALONE_HOME", ThirdPartySoftware.SPMSTANDALONE
        )
    )


def get_spm_home() -> Path:
    """Return the path to the home directory of SPM."""
    return check_environment_variable(
        SoftwareEnvironmentVariable("SPM_HOME", ThirdPartySoftware.SPM)
    )


def is_binary_present(name: str) -> bool:
    """Check if a binary is present.

    Parameters
    ----------
    name : str
        The name of the program.

    Returns
    -------
    True if the binary is present, False otherwise.

    Warnings
    --------
    Do not use this function with a binary GUI, it will open the GUI.

    References
    ----------
    Taken from:
    https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    Examples
    --------
    >>> is_binary_present("ls")
    True
    >>> is_binary_present("foo")
    False
    """
    import os
    import subprocess

    from clinica.compat import errno

    try:
        devnull = open(os.devnull)
        subprocess.Popen([name], stdout=devnull, stderr=devnull).communicate()
    except OSError as e:
        if e.errno == errno.ENOENT:
            return False
    return True


def check_binary(name: str):
    if not is_binary_present(name):
        raise ClinicaMissingDependencyError(
            f"Command {name} is unknown on your system. "
            "Please verify that you have correctly installed the corresponding software."
        )


def check_environment_variable(variable: SoftwareEnvironmentVariable) -> Path:
    """Check if the provided environment variable is set and returns its value.

    Parameters
    ----------
    variable : SoftwareEnvironmentVariable
        The software environment variable to check.

    Returns
    -------
    Path :
        The path value associated to the environment variable.

    Raises
    ------
    ClinicaMissingDependencyError :
        If the variable is not set.

    ClinicaEnvironmentVariableError :
        If the variable associated value is not a directory.

    Examples
    --------
    >>> check_environment_variable(SoftwareEnvironmentVariable("ANTSPATH", ThirdPartySoftware.ANTS))
    '/opt/ANTs/bin/'
    """
    from .exceptions import ClinicaEnvironmentVariableError

    if not (content_var := os.environ.get(variable.name, "")):
        raise ClinicaMissingDependencyError(
            f"Clinica could not find {variable.software.value} software: "
            f"the {variable.name} variable is not set."
        )
    if not os.path.isdir(content_var):
        raise ClinicaEnvironmentVariableError(
            f"The {variable.name} environment variable "
            f"you gave is not a folder (content: {content_var})."
        )
    return Path(content_var)


def _check_software(
    name: ThirdPartySoftware,
    binaries: Optional[Tuple[str]] = None,
    environment_variables: Optional[Tuple[SoftwareEnvironmentVariable]] = None,
    complementary_info: Optional[str] = None,
) -> None:
    """Check if the software is available.

    Parameters
    ----------
    name : ThirdPartySoftware
        Name of the software.

    binaries : tuple of str, optional
        Tuple of associated binaries to check.
        If None, nothing is checked.

    environment_variables : tuple of SoftwareEnvironmentVariable, optional
        These environment variables will be checked, meaning
        that they should be set and point to existing folders.
        If None, nothing is checked.

    complementary_info : str, optional
        Information specific to the software that should be
        added at the end of the error message when checking
        for binaries.

    Raises
    ------
    ClinicaMissingDependencyError
        If the software checks fail.
    """
    if environment_variables:
        for variable in environment_variables:
            check_environment_variable(variable)
    binaries = binaries or ()
    complementary_info = complementary_info or ""
    for binary in binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                f"[Error] Clinica could not find {name.value} software: "
                f"the {binary} command is not present in your PATH "
                f"environment. {complementary_info}"
            )


_check_dcm2niix = functools.partial(
    _check_software,
    name=ThirdPartySoftware.DCM2NIIX,
    binaries=("dcm2niix",),
    complementary_info=(
        "This software can be downloaded and installed "
        "from https://github.com/rordenlab/dcm2niix."
    ),
)

_check_ants = functools.partial(
    _check_software,
    name=ThirdPartySoftware.ANTS,
    binaries=("N4BiasFieldCorrection", "antsRegistrationSyNQuick.sh"),
)

_check_convert3d = functools.partial(
    _check_software,
    name=ThirdPartySoftware.CONVERT3D,
    binaries=("c3d_affine_tool", "c3d"),
)

_check_freesurfer = functools.partial(
    _check_software,
    name=ThirdPartySoftware.FREESURFER,
    binaries=("mri_convert", "recon-all"),
    environment_variables=(
        SoftwareEnvironmentVariable("FREESURFER_HOME", ThirdPartySoftware.FREESURFER),
    ),
    complementary_info=(
        "Do you have the line `source $FREESURFER_HOME/SetUpFreeSurfer.sh` "
        "in your configuration file?"
    ),
)

_check_mrtrix = functools.partial(
    _check_software,
    name=ThirdPartySoftware.MRTRIX,
    binaries=("transformconvert", "mrtransform", "dwi2response", "tckgen"),
)

_check_petpvc = functools.partial(
    _check_software,
    name=ThirdPartySoftware.PETPVC,
    binaries=(
        "petpvc",
        "pvc_diy",
        "pvc_gtm",
        "pvc_iy",
        "pvc_labbe",
        "pvc_make4d",
        "pvc_mg",
        "pvc_mtc",
        "pvc_rbv",
        "pvc_relabel",
        "pvc_rl",
        "pvc_simulate",
        "pvc_stc",
        "pvc_vc",
    ),
)


def _check_spm():
    """Check that SPM is installed, either regular with Matlab or as a standalone."""
    try:
        _check_spm_standalone()
    except ClinicaMissingDependencyError as e1:
        try:
            _check_spm_alone()
        except ClinicaMissingDependencyError as e2:
            raise ClinicaMissingDependencyError(
                "Clinica could not find the SPM software (regular or standalone).\n"
                "Please make sure you have installed SPM in one of these two ways "
                "and have set the required environment variables.\n"
                f"Full list of errors: \n- {e1}\n- {e2}"
            )


_check_spm_standalone = functools.partial(
    _check_software,
    name=ThirdPartySoftware.SPMSTANDALONE,
    environment_variables=(
        SoftwareEnvironmentVariable(
            "SPMSTANDALONE_HOME", ThirdPartySoftware.SPMSTANDALONE
        ),
        SoftwareEnvironmentVariable("MCR_HOME", ThirdPartySoftware.MCR),
    ),
)

_check_spm_alone = functools.partial(
    _check_software,
    name=ThirdPartySoftware.SPM,
    environment_variables=(
        SoftwareEnvironmentVariable("SPM_HOME", ThirdPartySoftware.SPM),
    ),
    binaries=("matlab",),
)

_check_matlab = functools.partial(
    _check_software,
    name=ThirdPartySoftware.MATLAB,
    binaries=("matlab",),
)

_check_fsl = functools.partial(
    _check_software,
    name=ThirdPartySoftware.FSL,
    binaries=("bet", "flirt", "fast", "first"),
    environment_variables=(
        SoftwareEnvironmentVariable("FSLDIR", ThirdPartySoftware.FSL),
    ),
)


def get_software_min_version_supported(
    software: Union[str, ThirdPartySoftware],
) -> Version:
    """Return the minimum version of the provided third-party software required by Clinica.

    Parameters
    ----------
    software : str or ThirdPartySoftware
        One of the third-party software of Clinica.

    Returns
    -------
    Version :
        The minimum version number of the software required by Clinica.

    Examples
    --------
    >>> from clinica.utils.check_dependency import get_software_min_version_supported
    >>> get_software_min_version_supported("ants")
    <Version('2.5.0')>
    """
    software = ThirdPartySoftware(software)
    if software == ThirdPartySoftware.FREESURFER:
        return Version("6.0.0")
    if software == ThirdPartySoftware.FSL:
        return Version("5.0.5")
    if software == ThirdPartySoftware.ANTS:
        return Version("2.5.0")
    if software == ThirdPartySoftware.DCM2NIIX:
        return Version("1.0.20240202")
    if software == ThirdPartySoftware.MRTRIX:
        return Version("3.0.3")
    if software == ThirdPartySoftware.CONVERT3D:
        return Version("1.0.0")
    if software == ThirdPartySoftware.MATLAB:
        return Version("9.2.0.556344")
    if software == ThirdPartySoftware.SPM:
        return Version("12.7219")
    if software == ThirdPartySoftware.MCR:
        return Version("9.0.1")
    if software == ThirdPartySoftware.SPMSTANDALONE:
        return Version("12.7219")
    if software == ThirdPartySoftware.PETPVC:
        return Version("0.0.0")


def get_software_version(software: Union[str, ThirdPartySoftware]) -> Version:
    """Return the version of the provided third-party software.

    Parameters
    ----------
    software : str or ThirdPartySoftware
        One of the third-party software of Clinica.

    Returns
    -------
    Version :
        The version number of the installed software.

    Notes
    -----
    This function assumes the software are correctly installed.
    It doesn't run any check and directly try to infer the version number by calling an
    underlying executable.

    Examples
    --------
    >>> from clinica.utils.check_dependency import get_software_version
    >>> get_software_version("freesurfer")
    <Version('7.2.0')>
    """
    software = ThirdPartySoftware(software)
    if software == ThirdPartySoftware.FREESURFER:
        return _get_freesurfer_version()
    if software == ThirdPartySoftware.FSL:
        return _get_fsl_version()
    if software == ThirdPartySoftware.ANTS:
        return _get_ants_version()
    if software == ThirdPartySoftware.DCM2NIIX:
        return _get_dcm2niix_version()
    if software == ThirdPartySoftware.MRTRIX:
        return _get_mrtrix_version()
    if software == ThirdPartySoftware.CONVERT3D:
        return _get_convert3d_version()
    if software == ThirdPartySoftware.MATLAB:
        return _get_matlab_version()
    if software == ThirdPartySoftware.SPM:
        return _get_spm_version()
    if software == ThirdPartySoftware.MCR:
        return _get_mcr_version()
    if software == ThirdPartySoftware.SPMSTANDALONE:
        return _get_spm_standalone_version()
    if software == ThirdPartySoftware.PETPVC:
        return Version("0.0.0")


def _get_freesurfer_version() -> Version:
    from nipype.interfaces import freesurfer

    return Version(str(freesurfer.Info.looseversion()))


def _get_spm_version() -> Version:
    from nipype.interfaces import spm

    from .spm import configure_nipype_interface_to_work_with_spm

    configure_nipype_interface_to_work_with_spm()

    return Version(spm.SPMCommand().version)


def _get_spm_standalone_version() -> Version:
    from nipype.interfaces import spm

    from .spm import configure_nipype_interface_to_work_with_spm_standalone

    configure_nipype_interface_to_work_with_spm_standalone()

    return Version(spm.SPMCommand().version)


def _get_fsl_version() -> Version:
    import re

    from nipype.interfaces import fsl

    from clinica.utils.stream import cprint

    raw_output = str(fsl.Info.version())
    try:
        return Version(re.search(r"\s*([\d.]+)", raw_output).group(1))
    except Exception as e:
        cprint(msg=str(e), lvl="error")


def _get_software_version_from_command_line(
    executable: str, prepend_with_v: bool = True, two_dashes: bool = True
) -> Version:
    import re

    from clinica.utils.stream import log_and_raise

    try:
        return Version(
            re.search(
                rf"{'v' if prepend_with_v else ''}\s*([\d.]+)",
                _run_command(executable, two_dashes=two_dashes),
            )
            .group(1)
            .strip(".")
        )
    except Exception as e:
        log_and_raise(str(e), RuntimeError)


def _run_command(executable: str, two_dashes: bool = True) -> str:
    import subprocess

    return subprocess.run(
        [executable, f"-{'-' if two_dashes else ''}version"], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")


_get_ants_version = partial(
    _get_software_version_from_command_line,
    executable="antsRegistration",
    prepend_with_v=False,
)
_get_dcm2niix_version = partial(
    _get_software_version_from_command_line, executable="dcm2niix"
)
_get_mrtrix_version = partial(
    _get_software_version_from_command_line,
    executable="mrtransform",
    prepend_with_v=False,
)
_get_convert3d_version = partial(
    _get_software_version_from_command_line,
    executable="c3d",
    prepend_with_v=False,
    two_dashes=False,
)


def _get_matlab_version() -> Version:
    import re

    from clinica.utils.stream import log_and_raise

    try:
        return Version(
            re.search(r"\(\s*([\d.]+)\)", _get_matlab_start_session_message())
            .group(1)
            .strip(".")
        )
    except Exception as e:
        log_and_raise(str(e), RuntimeError)


def _get_matlab_start_session_message() -> str:
    """Start Matlab, get the message displayed at the beginning of the session, and quit."""
    import subprocess

    return subprocess.run(
        ["matlab", "-r", "quit", "-nojvm"], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")


def _get_mcr_version() -> Version:
    import os

    mcr_home_path = Path(os.environ.get("MCR_HOME"))
    raw_version = mcr_home_path.parent.name
    return _map_mcr_release_to_version_number(raw_version)


def _map_mcr_release_to_version_number(mcr_release: str) -> Version:
    """Map the MCR release to a proper version number.

    If the found version is older than the minimum supported version, we don't bother mapping it
    to its version number, 0.0.0 is returned.
    If the release is >= 2024a, we use the fact that Matlab is now using proper calendar versioning.
    If the release is in-between, we use a hardcoded mapping.
    """
    mcr_versions_mapping = {
        "2023b": Version("23.2"),
        "2023a": Version("9.14"),
        "2022b": Version("9.13"),
        "2022a": Version("9.12"),
        "2021b": Version("9.11"),
        "2021a": Version("9.10"),
        "2020b": Version("9.9"),
        "2020a": Version("9.8"),
        "2019b": Version("9.7"),
        "2019a": Version("9.6"),
        "2018b": Version("9.5"),
        "2018a": Version("9.4"),
        "2017b": Version("9.3"),
        "2017a": Version("9.2"),
        "2016b": Version("9.1"),
        "2016a": Version("9.0.1"),
    }
    if int(mcr_release[:4]) >= 2024:
        return Version(f"{mcr_release[2:4]}.{'1' if mcr_release[-1] == 'a' else '2'}")
    return mcr_versions_mapping.get(mcr_release, Version("0.0.0"))


def _check_software_version(
    software: ThirdPartySoftware,
    *,
    log_level: Optional[LoggingLevel] = None,
    specifier: Optional[SpecifierSet] = None,
):
    """Check that the installed version of the software is satisfying the constraints imposed by Clinica."""
    log_level = log_level or LoggingLevel.WARNING
    if specifier is None:
        specifier = SpecifierSet(f">={get_software_min_version_supported(software)}")
    if (installed_version := get_software_version(software)) not in specifier:
        (log_and_raise if log_level >= LoggingLevel.ERROR else log_and_warn)(
            f"{software.value} version is {installed_version}. We strongly recommend to have {software.value} {specifier}.",
            (
                ClinicaMissingDependencyError
                if log_level == LoggingLevel.ERROR
                else UserWarning
            ),
        )
    cprint(
        f"Found installation of {software.value} with version {installed_version}, satisfying {specifier}.",
        lvl="info",
    )


def check_software(
    software: Union[str, ThirdPartySoftware],
    *,
    log_level: Optional[Union[str, LoggingLevel]] = None,
    specifier: Optional[Union[str, SpecifierSet]] = None,
):
    """Run some checks on the given software.

    These checks are of two types:
        - checks to verify that the executable is present in the PATH.
          Also check configurations made through environment variables.
        - checks on the installed version. The installed version has to
          satisfy the constraints imposed by Clinica.

    Parameters
    ----------
    software : str or ThirdPartySoftware
        One of the third-party software of Clinica.

    log_level : str or LoggingLevel, optional
        Whether to raise or warn if the version of the installed software does not
        match the requirement. For specific cases like if FreeSurfer is < 6.0, an
        error will be raised independently of this parameter because it is impossible
        that Clinica pipelines could work in these situations.
        Default is set to warnings.

    specifier : str or SpecifierSet, optional
        A version constraint for the software (i.e. '>=2.0.1', '<1.0.0', or '==0.10.9').
        If not provided, the specifier will be '>= minimum_version' where 'minimum_version'
        is the minimum version number of the software required by Clinica.

    Raises
    ------
    ClinicaMissingDependencyError :
        If an issue is found with the installation of the provided software.
        In some cases where the software is correctly installed, but the version is considered
        a bit old compared to the one recommended by Clinica, a warning could be given instead of
        an error. In this situation, it is strongly recommended to upgrade the dependency even
        though Clinica does not raise.

    Examples
    --------
    >>> from clinica.utils.check_dependency import check_software
    >>> check_software("dcm2niix")
    >>> check_software("matlab", log_level="error")
    >>> check_software("ants", specifier=">=2.5")
    """
    software = ThirdPartySoftware(software)
    if specifier and not isinstance(specifier, SpecifierSet):
        if specifier == "":
            specifier = None
        else:
            specifier = SpecifierSet(specifier)
    if log_level:
        log_level = get_logging_level(log_level)
    else:
        log_level = LoggingLevel.WARNING
    if software == ThirdPartySoftware.ANTS:
        _check_ants()
    if software == ThirdPartySoftware.FSL:
        _check_fsl()
        log_level = LoggingLevel.ERROR
    if software == ThirdPartySoftware.FREESURFER:
        _check_freesurfer()
        log_level = LoggingLevel.ERROR
    if (
        software == ThirdPartySoftware.SPM
        or software == ThirdPartySoftware.SPMSTANDALONE
        or software == ThirdPartySoftware.MCR
    ):
        _check_spm()
        log_level = LoggingLevel.ERROR
    if software == ThirdPartySoftware.MATLAB:
        _check_matlab()
    if software == ThirdPartySoftware.DCM2NIIX:
        _check_dcm2niix()
    if software == ThirdPartySoftware.PETPVC:
        _check_petpvc()
    if software == ThirdPartySoftware.MRTRIX:
        _check_mrtrix()
    if software == ThirdPartySoftware.CONVERT3D:
        _check_convert3d()
    _check_software_version(software, log_level=log_level, specifier=specifier)



from typing import Iterable, Union

from pydra.mark import annotate, task


@task
@annotate({"return": {"prepared_flowfields": list}})
def prepare_flowfields_task(flowfields: Iterable, number_of_tissues: int) -> list:
    """Pydra task for prepare_flow_fields."""
    return prepare_flowfields(flowfields, number_of_tissues)


def prepare_flowfields(flowfields: Iterable, number_of_tissues: int) -> list:
    """Reshape the flow fields according to the number of tissues.

    This function broadcasts the flow fields to the number of tissues
    selected by the user.

    Parameters
    ----------
    flowfields : list of str or str
        The path to the flowfields. It can be a string if we are dealing
        with a path to a single file, or a list if we are dealing with a
        list of paths.

    number_of_tissues : int
        The number of tissues selected.

    Returns
    -------
    list:
        Either a simple list with the flowfields path repeated for each tissue,
        of a list of lists of shape (n_flowfields, n_tissues).

    Raises
    ------
    TypeError
        If the provided flowfields is not a string, a list, or a tuple.
        If the provided number of tissues is not an integer.

    Examples
    --------
    >>> prepare_flowfields(["path_to_flowfield_1", "path_to_flowfield_2"], 3)
    [['path_to_flowfield_1', 'path_to_flowfield_1', 'path_to_flowfield_1'], ['path_to_flowfield_2', 'path_to_flowfield_2', 'path_to_flowfield_2']]
    >>> prepare_flowfields(["path_to_flowfield_1", "path_to_flowfield_2"], 2)
    [['path_to_flowfield_1', 'path_to_flowfield_1'], ['path_to_flowfield_2', 'path_to_flowfield_2']]
    >>> prepare_flowfields(["path_to_flowfield_1"], 2)
    [['path_to_flowfield_1', 'path_to_flowfield_1']]
    >>> prepare_flowfields("path_to_flowfield_1", 2)
    ['path_to_flowfield_1', 'path_to_flowfield_1']
    """
    if not isinstance(number_of_tissues, int):
        raise TypeError(
            f"Invalid type for number_of_tissues. Expected int, got {type(number_of_tissues)}."
        )
    if isinstance(flowfields, (list, tuple)):
        return [[f] * number_of_tissues for f in flowfields]
    if isinstance(flowfields, str):
        return [flowfields] * number_of_tissues
    raise TypeError(f"Invalid flowfields type: {type(flowfields)}.")



"""This module contains utilities common to all DWI pipelines."""

from collections import namedtuple
from pathlib import Path
from typing import Dict, Tuple

__all__ = ["DWIDataset", "rename_files"]


DWIDataset = namedtuple("DWIDataset", "dwi b_values b_vectors")


def rename_files(in_caps_dwi: Path, mapping: Dict[Path, str]) -> Tuple[str, ...]:
    """Rename files provided.

    The new files are symbolic links to old files.
    For this reason, the old files still exists after renaming.

    Parameters
    ----------
    in_caps_dwi : Path
        A DWI file from the CAPS folder.
        This is used only to extract the BIDS identifier.

    mapping : dict<Path, str>
        Mapping between original file paths and suffixes for
        new file names.

    Returns
    -------
    tuple :
        New file names.
    """
    from nipype.interfaces.utility import Rename

    bids_id = _extract_bids_identifier_from_filename(str(in_caps_dwi))
    renamed_files = []
    for original_file, suffix in mapping.items():
        rename = Rename()
        rename.inputs.in_file = str(original_file)
        rename.inputs.format_string = str(original_file.parent / f"{bids_id}{suffix}")
        renamed_files.append(rename.run().outputs.out_file)

    return tuple(renamed_files)


def _extract_bids_identifier_from_filename(caps_dwi_filename: str) -> str:
    """Extract BIDS identifier from a DWI CAPS filename.

    Parameters
    ----------
    caps_dwi_filename : str
        DWI file name for which to extract the bids identifier.

    Returns
    -------
    str :
        The corresponding BIDS identifier.

    Examples
    --------
    >>> _extract_bids_identifier_from_filename("sub-01_ses-M000_dwi_space-b0_preproc.bval")
    'sub-01_ses-M000'
    >>> _extract_bids_identifier_from_filename("sub-01_ses-M000_dwi.bvec")
    'sub-01_ses-M000'
    >>> _extract_bids_identifier_from_filename("foo/bar/sub-01_ses-M000_dwi_baz.foo.bar")
    'sub-01_ses-M000'
    >>> _extract_bids_identifier_from_filename("foo/bar/sub-01_ses-M000_space-b0_desc-preproc_dwi.bval")
    'sub-01_ses-M000_space-b0_desc-preproc'
    """
    import re

    m = re.search(r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+).*_dwi", caps_dwi_filename)
    if not m:
        raise ValueError(
            f"Could not extract the BIDS identifier from the DWI input filename {caps_dwi_filename}."
        )
    identifier = m.group(0).rstrip("_dwi")

    return identifier



"""This module contains utilities for statistics.

Currently, it contains one function to generate TSV file containing mean map based on a parcellation.
"""
from os import PathLike
from pathlib import Path
from typing import Optional, Union

from clinica.utils.atlas import AtlasName, BaseAtlas


def statistics_on_atlas(
    in_normalized_map: Union[str, PathLike],
    atlas: Union[str, AtlasName, BaseAtlas],
    out_file: Optional[Union[str, PathLike]] = None,
) -> str:
    """Compute statistics of a map on an atlas.

    Given an atlas image with a set of ROIs, this function computes the mean of
    a normalized map (e.g. GM segmentation, FA map from DTI, etc.) on each ROI.

    Parameters
    ----------
    in_normalized_map : str
        File containing a scalar image registered on the atlas.

    atlas : BaseAtlas or AtlasName or str
        An atlas with a set of ROI. These ROI are used to compute statistics.
        If a string is given, it is assumed to be the name of the atlas to be used.

    out_file : str, optional
        Name of the output file.

    Returns
    -------
    out_file : str
        TSV file containing the statistics (content of the columns: label,
        mean scalar, std of the scalar', number of voxels).
    """
    import nibabel as nib
    import numpy as np
    import pandas as pd

    from clinica.utils.stream import cprint

    from .atlas import atlas_factory

    atlas = atlas_factory(atlas)
    in_normalized_map = Path(in_normalized_map)
    if not out_file:
        filename, ext = in_normalized_map.stem, in_normalized_map.suffix
        if ext == ".gz":
            filename = Path(filename).stem
        out_file = Path(f"{filename}_statistics_{atlas.name}.tsv").resolve()

    atlas_labels = nib.load(atlas.labels)
    atlas_labels_data = atlas_labels.get_fdata(dtype="float32")

    img = nib.load(in_normalized_map)
    img_data = img.get_fdata(dtype="float32")

    atlas_correspondence = pd.read_csv(atlas.tsv_roi, sep="\t")
    label_name = list(atlas_correspondence.roi_name)
    # TODO create roi_value column in lut_*.txt and remove irrelevant RGB information
    label_value = list(atlas_correspondence.roi_value)

    mean_signal_value = []
    for label in label_value:
        current_mask_label = atlas_labels_data == label
        masked_data = np.array(img_data, copy=True)
        masked_data[np.invert(current_mask_label)] = 0
        mean_signal_value.append(np.sum(masked_data) / np.sum(current_mask_label))

    try:
        data = pd.DataFrame(
            {"label_name": label_name, "mean_scalar": mean_signal_value}
        )
        data.to_csv(out_file, sep="\t", index=True, encoding="utf-8")
    except Exception as e:
        cprint(msg=f"Impossible to save {out_file} with pandas", lvl="error")
        raise e

    return out_file
